"""Declares :class:`NamedQueryArgs`."""
import pydantic


class NamedQueryArgs(pydantic.BaseModel):
    """A :class:`pydantic.BaseModel` implementation that provides some hooks
    for the :class:`cbra.QueryRunner` to execute the query.
    """



"""Declares :func:`RFC9068Principal`."""
import inspect
import typing

import fastapi
import httpx
from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import JSONWebKeySet
from ckms.types import ServerMetadata
from fastapi.security import HTTPBearer
from fastapi.security import HTTPAuthorizationCredentials

from cbra.params import ServerKeychain
from cbra.exceptions import AuthenticationRequired
from cbra.exceptions import CanonicalException
from cbra.exceptions import ForgedAccessToken
from cbra.exceptions import InvalidAuthorizationScheme
from cbra.exceptions import UntrustedIssuer
from cbra.utils import retry
from .principal import Principal
from .types import IPrincipal
from .params import LocalIssuer


security = HTTPBearer(auto_error=False)
ISSUER_JWKS_CACHE: dict[str, JSONWebKeySet] = {}


@retry(5, interval=2.0)
async def get_issuer_jwks(issuer: str) -> JSONWebKeySet | None:
    jwks = ISSUER_JWKS_CACHE.get(issuer)
    if jwks is None:
        async with httpx.AsyncClient() as client:
            metadata = await ServerMetadata.discover(client, issuer)
            jwks = await metadata.get_jwks(client)
        if jwks is not None:
            ISSUER_JWKS_CACHE[issuer] = jwks
    return jwks


def RFC9068Principal(
    auto_error: bool = True,
    client_id: typing.Optional[str] = None,
    trusted_issuers: typing.Optional[typing.Set[str]] = None,
    header: str = 'Authorization',
    path: typing.Optional[str] = None,
    max_age: int = 300,
    principal_factory: typing.Callable[[typing.Any], IPrincipal] = Principal.fromclaimset,
    scope: set[str] | None = None,
    null: type[IPrincipal] | None = None
) -> typing.Any:
    """Resolves the principal in a request to a subject using an
    :rfc:`9068` access token.
    """
    scope = scope or set()
    issuers: typing.Set[str] = trusted_issuers or set()
    if path is not None and not str.startswith(path, '/'):
        raise ValueError("The `path` parameter must start with a slash.")

    async def resolve_principal(
        request: fastapi.Request,
        issuer: str = LocalIssuer,
        bearer: typing.Optional[
            HTTPAuthorizationCredentials
        ] = fastapi.Depends(security),
        keychain: Keychain = ServerKeychain
    ):
        codec = PayloadCodec()
        if bearer is None:
            if auto_error and 'Authorization' not in request.headers:
                raise AuthenticationRequired
            if auto_error:
                # The header was present but a parsing error occurred.
                raise InvalidAuthorizationScheme
            return null or None

        if str.lower(bearer.scheme or '') != "bearer":
            raise InvalidAuthorizationScheme

        audience = {
            f"{request.url.scheme}://{request.url.netloc}",
            f"{request.url.scheme}://{request.url.netloc}{path or ''}",
            f"{request.url.scheme}://{request.url.netloc}{request.url.path}"
        }
        jws, claims = await codec.jwt(bearer.credentials, accept="at+jwt")
        claims.verify(
            audience=audience,
            required={"jti", "iss", "aud", "sub", "iat", "nbf", "exp", "client_id"},
            max_age=max_age
        )
        if claims.iss not in (issuers | {issuer}):
            assert claims.iss is not None # nosec
            raise UntrustedIssuer(claims.iss)

        granted_scope: set[str] = {
            x for x in str.split((claims.extra.get('scope') or ''), ' ')
            if x
        }
        if not (granted_scope >= scope):
            raise CanonicalException(
                http_status_code=401,
                code="AUTHORIZATION_REQUIRED",
                message=(
                    "The scope granted to the access token was not sufficient "
                    "to access this resource."
                )
            )

        # If the issuer was valid but the signature validation fails here,
        # then the token was most probably forged.
        if claims.iss == issuer:
            is_valid_signature = jws.verify(keychain, require_kid=True)
        else:
            # Token was not issued by this server. The issuer must then be in the
            # trusted issuers set.
            assert claims.iss in issuers # nosec
            jwks = await get_issuer_jwks(claims.iss)
            if jwks is None:
                raise CanonicalException(
                    http_status_code=503,
                    code="ISSUER_UNREACHABLE",
                    message=(
                        "Unable to reach the issuer specified by the access token."
                    )
                )
            is_valid_signature = jws.verify(jwks, require_kid=True)
        if inspect.isawaitable(is_valid_signature):
            is_valid_signature = await is_valid_signature
        if not is_valid_signature:
            raise ForgedAccessToken
        return principal_factory(claims)

    return resolve_principal


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`EndpointMetaclass`."""
from typing import Any


class EndpointMetaclass(type):
    __module__: str = 'cbra'

    def __new__(
        cls,
        name: str,
        bases: tuple[type[object]],
        attrs: dict[str, Any]
    ):
        """Construct a new :class:`~cbra.Endpoint` class."""
        super_new = super().__new__
        if attrs.pop('__abstract__', False):
            return super_new(cls, name, bases, attrs)
        new_class = super_new(cls, name, bases, attrs)

        return new_class


"""Declares :class:`IQueryRunner`."""
import functools
from typing import Any


class IQueryRunner:
    """Specifies an abstract interface to run read-only data queryies."""
    __module__: str = 'ckms.types'

    @functools.singledispatchmethod
    async def execute(self, query: Any) -> Any:
        """Execute the `query` with the specified parameters."""
        raise NotImplementedError(
            f"No handler was defined for {type(query).__name__}"
        )



"""Declares :class:`RequestHandler`."""
import abc
import datetime
import typing

import pydantic

from .openauthorizationendpoint import OpenAuthorizationEndpoint


class RequestHandler(metaclass=abc.ABCMeta):
    __module__: str = 'cbra.ext.oauth2'

    #: A callable that returns a :class:`pydantic.BaseModel` holding the
    #: response parameters.
    response_model: typing.Optional[typing.Type[pydantic.BaseModel]] = None

    #: Autoritative date for the request.
    now: datetime.datetime

    @classmethod
    def as_endpoint(
        cls,
        **kwargs: typing.Any
    ) -> OpenAuthorizationEndpoint:
        """Return a callable that handles an incoming request to the OAuth 2.0
        handler.
        """
        return OpenAuthorizationEndpoint(cls, **kwargs)

    def _setup(self, **kwargs: typing.Any):
        self.now = datetime.datetime.utcnow()
        self.setup(**kwargs)

    def setup(self, *args: typing.Any, **kwargs: typing.Any):
        """Hook to configure a handler instance. This method received the
        keyword arguments that were passed to :meth:`as_endpoint()`.
        """
        pass


"""Declares :class:`JSONRenderer`."""
import json
import typing

import pydantic

from cbra.encoders import JSONEncoder
from cbra.types import IRenderer
from cbra.utils import parse_header


class JSONRenderer(IRenderer):
    """Renderer which serializes to JSON."""
    __module__: str = 'cbra.renderers'
    charset: typing.Optional[str] = None
    compact: bool = True
    encoder_class: type[JSONEncoder] = JSONEncoder
    ensure_ascii: bool = True
    format: str = 'json'
    indent: typing.Optional[int] = None
    media_type: str = 'application/json'
    response_media_type: str = "application/json"
    strict: bool = True
    structured: bool = True

    def has_content(self) -> bool:
        return True

    def get_indent(
        self,
        accepted_media_type: str,
        renderer_context: typing.Dict[str, typing.Any]
    ) -> typing.Optional[int]:
        if self.accepted:
            _, params = parse_header(self.accepted.encode('ascii'))
            try:
                indent = typing.cast(int, params['indent']) # type: ignore
                v = max(min(int(indent), 8), 0)
                return None if v == 0 else v
            except (KeyError, ValueError, TypeError):
                pass

        # If 'indent' is provided in the context, then pretty print the result.
        return renderer_context.get('indent', None) or self.indent

    def render(
        self,
        data: dict[str, typing.Any] | pydantic.BaseModel,
        renderer_context: dict[str, typing.Any] | None = None
    ) -> bytes:
        """Render `data` into JSON, returning a bytestring."""
        if data is None:
            return b''

        renderer_context = renderer_context or {}
        indent = self.get_indent(self.accepted, renderer_context)
        if indent is None:
            separators = (',', ':')
        else:
            separators = (', ', ': ')

        ret = None
        if isinstance(data, pydantic.BaseModel):
            ret = data.json(
                indent=indent,
                ensure_ascii=self.ensure_ascii,
                allow_nan=not self.strict,
                separators=separators,
                by_alias=renderer_context.get('by_alias'),
                exclude=renderer_context.get('exclude'),
                exclude_defaults=renderer_context.get('exclude_defaults'),
                exclude_none=renderer_context.get('exclude_none'),
            )
        else:
            ret = json.dumps(
                data,
                cls=self.encoder_class,
                indent=indent,
                ensure_ascii=self.ensure_ascii,
                allow_nan=not self.strict,
                separators=separators
            )

        # We always fully escape \u2028 and \u2029 to ensure we output JSON
        # that is a strict javascript subset.
        # See: https://gist.github.com/damncabbage/623b879af56f850a6ddc
        ret = ret.replace('\u2028', '\\u2028').replace('\u2029', '\\u2029')
        return ret.encode()



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Webhooks are an integral part of modern web applications. Many applications
consume external services through (REStful) APIs, and those services often
provide a mechanism to deliver event notifications to their consumers: webhooks.
The :mod:`cbra` framework provides a class-oriented solution to handle these
incoming messages.
"""
from typing import cast
from typing import Callable
from types import UnionType

import pydantic

from . import models
from .webhookendpoint import WebhookEndpoint


__all__ = [
    'models',
    'WebhookEndpoint'
]


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re

import pydantic
from ckms.types import JSONWebToken
from ckms.types import MalformedPayload

from cbra.ext.oauth2.types import IPrincipal
from .clientprincipal import ClientPrincipal
from .subjectprincipal import SubjectPrincipal


class AccessTokenPrincipal(pydantic.BaseModel, IPrincipal):
    __root__: ClientPrincipal | SubjectPrincipal

    def has_scope(
        self,
        scope: str | set[str]
    ) -> bool:
        """Return a boolean indicating if the :class:`AccessTokenPrincipal` is
        authorized to use the given `scope`.
        """
        if isinstance(scope, str):
            scope = {scope}
        return scope <= self.__root__.scope

    @classmethod
    def fromclaimset(cls, claims: JSONWebToken) -> 'AccessTokenPrincipal':
        assert claims.sub is not None # nosec
        scope = []
        if 'scope' in claims.extra:
            if not isinstance(claims.extra['scope'], str):
                raise MalformedPayload(
                    message="The 'scope' parameter must be a space-separated string."
                )
            scope = re.split('\\s', claims.extra['scope'])
        return cls.parse_obj({
            **claims.dict(),
            'scope': scope
        })

    def is_authenticated(self) -> bool:
        return self.__root__.is_authenticated()


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi


class TokenException(Exception):
    __module__: str = 'cbra.ext.oauth2.types'
    code: str
    description: str | None = None

    def __init__(self, code: str, description: str | None = None):
        self.code = code
        self.description = description

    def as_response(self) -> fastapi.responses.JSONResponse:
        return fastapi.responses.JSONResponse(
            status_code=400,
            content={
                'error': self.code,
                'error_description': self.description or (
                    "The server refuses to disclose any further information regarding "
                    "the error condition."
                )
            }
        )


"""Declares :class:`IStorage`."""
import functools
import typing
from typing import Any

from ckms.types import ClaimSet

from ..exceptions import Error
from .authorizationcode import AuthorizationCode
from .authorizationidentifier import AuthorizationIdentifier
from .authorizationrequest import AuthorizationRequest
from .baseauthorizationrequest import BaseAuthorizationRequest
from .iauthorization import IAuthorization
from .irefreshtoken import IRefreshToken
from .refreshtokenidentifier import RefreshTokenIdentifier


class IStorage:
    """Declares the interface that storage implementations for the OAuth 2.0
    server need to implement.
    """
    __module__: str = 'cbra.ext.oauth2.types'
    InvalidAuthorizationCode: type[Error] = type('InvalidAuthorizationCode', (Error, ), {
        'error': "invalid_grant",
        'error_description': "The presented authorization code grant is not valid."
    })

    async def consume(self, claims: ClaimSet) -> bool:
        """Consume the claims set, ensuring that its token can only be
        used once. Return a boolean indicating if the token was already
        consumed.
        """
        raise NotImplementedError

    async def delete_authorization_request(self, request: Any) -> None:
        """Delete the provided authorization request"""
        raise NotImplementedError

    @functools.singledispatchmethod
    async def get(
        self,
        key: AuthorizationIdentifier | RefreshTokenIdentifier
    ) -> Any:
        """Return a key from the transient storage backend, or ``None``
        if it does not exist.
        """
        raise NotImplementedError

    @get.register
    async def _get_authorization_by_natural_key(
        self,
        key: AuthorizationIdentifier
    ) -> Any:
        return await self.get_authorization_by_natural_key(key.client_id, key.sub)

    @get.register
    async def _get_refresh_token_by_id(
        self,
        key: RefreshTokenIdentifier
    ) -> IRefreshToken | None:
        return await self.get_refresh_token_by_id(key)

    async def get_authorization_by_natural_key(
        self,
        client_id: int | str,
        subject_id: int | str
    ) -> Any:
        raise NotImplementedError

    async def get_authorization_request(
        self,
        request_id: str
    ) -> AuthorizationRequest:
        """Lookup an authorization request from the transient storage."""
        raise NotImplementedError

    async def get_code(self, code: str) -> tuple[AuthorizationCode, AuthorizationRequest]:
        """Lookup an authorization code."""
        raise NotImplementedError

    async def get_pushed_authorization_request(
        self,
        request_uri: str
    ) -> BaseAuthorizationRequest:
        """Look an authorization request that was pushed through the Pushed
        Authorization Request (PAR) endpoint.
        """
        raise NotImplementedError

    async def get_refresh_token_by_id(
        self,
        key: RefreshTokenIdentifier
    ) -> IRefreshToken | None:
        """Return the :class:`~cbra.ext.oauth2.types.IRefreshToken` by its
        identifier `key`.
        """
        raise NotImplementedError

    async def set(
        self,
        key: str,
        value: typing.Union[bytes, str],
        expires: typing.Optional[int] = None
    ) -> None:
        """Set a `key` to given `value` to the **transient** storage backend.
        If `value` is a :class:`dict`, then all its members must be
        JSON-serializable.

        The `expires` parameter define the period after which the key expires
        i.e. it should be (automatically) removed from the transient storage
        backend. It is denoted in seconds.
        """
        raise NotImplementedError

    @functools.singledispatchmethod
    async def persist(
        self,
        obj: AuthorizationRequest | AuthorizationCode | IAuthorization | IRefreshToken
    ) -> None:
        """Persist an object to the transient storage."""
        raise NotImplementedError(type(obj).__name__)

    @persist.register
    async def _persist_authorization(
        self,
        obj: IAuthorization
    ) -> None:
        return await self.persist_authorization(obj)

    @persist.register
    async def _persist_authorizationcode(
        self,
        obj: AuthorizationCode
    ) -> None:
        return await self.persist_authorizationcode(obj)

    @persist.register
    async def _persist_authorizationrequest(
        self,
        obj: AuthorizationRequest
    ) -> None:
        return await self.persist_authorizationrequest(obj)

    @persist.register
    async def _persist_refresh_token(
        self,
        obj: IRefreshToken
    ) -> None:
        return await self.persist_refresh_token(obj)

    async def persist_authorization(
        self,
        obj: IAuthorization
    ) -> None:
        raise NotImplementedError

    async def persist_authorizationcode(
        self,
        obj: AuthorizationCode
    ) -> None:
        """Persist an :class:`cbra.ext.oauth2.types.AuthorizationCode`
        instance.
        """
        raise NotImplementedError

    async def persist_authorizationrequest(
        self,
        obj: AuthorizationRequest
    ) -> None:
        """Persist an :class:`cbra.ext.oauth2.AuthorizationRequest`
        instance.
        """
        raise NotImplementedError

    async def persist_refresh_token(
        self,
        obj: IRefreshToken
    ) -> None:
        """Persist an :class:`cbra.ext.oauth2.types.IRefreshToken`
        instance.
        """
        raise NotImplementedError

    async def pop(
        self,
        key: str
    ) -> typing.Optional[typing.Union[bytes, str]]:
        """Like :meth:`get()`, but also removes the key."""
        raise NotImplementedError


"""Declares :class:`IRequestHandler`."""
import typing

import fastapi


class IRouteable:
    """The interface for objects that add routes to an ASGI
    application.
    """
    __module__: str = 'cbra.types'
    RouterType: typing.TypeAlias = typing.Union[
        fastapi.FastAPI,
        fastapi.APIRouter
    ]

    @classmethod
    def add_to_router(
        cls,
        *,
        app: RouterType,
        base_path: str,
        default_error_responses: bool = True,
        **kwargs: typing.Any
    ) -> None:
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets

import httpx
from ckms.core import KeySpecification
from ckms.jose import PayloadCodec
from ckms.utils import current_timestamp

from cbra.ext.oauth2.types import TokenResponse
from .accesstoken import AccessTokenCredential
from .clientcredential import ClientCredential


class JWTPrivateKeyClientCredential(ClientCredential):
    """The ``private_key_jwt`` authentication method used by the token
    endpoint.
    """
    __module__: str = 'cbra.ext.api.credential'
    access_token: AccessTokenCredential | None
    client_assertion_type: str = 'urn:ietf:params:oauth:client-assertion-type:jwt-bearer'
    codec: PayloadCodec
    encryption_key: KeySpecification | None
    signing_key: KeySpecification
    token_endpoint: str

    def __init__(
        self,
        client_id: str,
        token_endpoint: str,
        signing_key: KeySpecification,
        encryption_key: KeySpecification | None
    ):
        super().__init__(client_id=client_id)
        self.access_token = None
        self.client_id = client_id
        self.encryption_key = encryption_key
        self.signing_key = signing_key
        self.token_endpoint = token_endpoint
        self.codec = PayloadCodec(
            signing_keys=[self.signing_key],
        )

    async def obtain(self, audience: str) -> AccessTokenCredential:
        """Obtain an access token using the client credentials grant."""
        params: dict[str, str] = {
            'client_id': self.client_id,
            'client_assertion_type': self.client_assertion_type,
            'client_assertion': await self.create_assertion(sub=self.client_id),
            'grant_type': 'client_credentials',
            'resource': audience
        }
        async with httpx.AsyncClient() as client:
            response = await client.post(url=self.token_endpoint, json=params) # type: ignore
            if response.status_code != 200:
                raise Exception(response.content)
            result = TokenResponse.parse_obj(response.json())
        return AccessTokenCredential(
            token=result.access_token,
            expires_in=result.expires_in,
            refresh_token=result.refresh_token
        )

    async def create_assertion(self, **claims: int | str) -> str:
        """Create the assertion used to authenticate with the token endpoint."""
        return await self.codec.encode(
            payload=self.get_claims(**claims)
        )

    def get_claims(self, **extra: int | str) -> dict[str, int | str]:
        """Return a dictionary containing the JWT claims used to
        construct the client assertion or JAR.
        """
        now = current_timestamp()
        return {
            **extra,
            'aud': self.token_endpoint,
            'client_id': self.client_id,
            'exp': now + 60,
            'iat': now,
            'iss': self.client_id,
            'jti': secrets.token_urlsafe(48),
            'nbf': now
        }


# pylint: skip-file
import asyncio
from typing import AsyncGenerator

import aorta.transport
import pytest
import pytest_asyncio
import ckms.core
from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import IKeySpecification
from httpx import AsyncClient

from cbra import Application
from cbra.ext import ioc
from cbra.ext.oauth2 import AccessTokenSigner


@pytest.fixture(scope='session') # type: ignore
def base_url() -> str:
    return "https://cbra.localhost"


@pytest_asyncio.fixture # type: ignore
async def client(
    app: Application,
    base_url: str
) -> AsyncGenerator[AsyncClient, None]:
    await app.on_startup()
    params = {
        'app': app,
        'base_url': base_url
    }
    async with AsyncClient(**params) as client:
        yield client


@pytest.fixture # type: ignore
def codec(app: Application) -> PayloadCodec:
    return app.codec


@pytest.fixture(scope="session") # type: ignore
def event_loop():
    return asyncio.new_event_loop()


@pytest_asyncio.fixture(scope='session') # type: ignore
async def jose_client_signer() -> IKeySpecification:
    return await ckms.core.parse_spec({
        'provider': 'local',
        'kty': 'oct',
        'algorithm': 'HS256',
        'key': {'length': 32}
    })


@pytest.fixture(autouse=True, scope='session') # type: ignore
def inject_message_publisher():
    ioc.provide('MessageTransport', aorta.transport.NullTransport())


@pytest_asyncio.fixture(scope='session') # type: ignore
async def server_keychain() -> Keychain:
    k = Keychain()
    k.configure(
        keys={
            'sig': {
                'provider': 'local',
                'kty': 'OKP',
                'alg': 'EdDSA',
                'crv': 'Ed448',
                'key': {'path': 'pki/sig.key'},
                'use': 'sig'
            },
        }
    )
    await k
    return k

@pytest.fixture # type: ignore
def signing_key() -> str:
    return 'sig'


@pytest.fixture # type: ignore
def token_signer(
    base_url: str,
    server_keychain: Keychain
) -> AccessTokenSigner:
    return AccessTokenSigner(
        issuer=base_url,
        codec=PayloadCodec(
            signer=server_keychain,
            encrypter=server_keychain
        )
    )


@pytest_asyncio.fixture(scope='session') # type: ignore
async def unknown_keychain() -> Keychain:
    k = Keychain()
    k.configure(
        keys={
            'sig': {
                'provider': 'local',
                'kty': 'RSA',
                'key': {'path': 'pki/ext-sig.key'},
                'use': 'sig'
            },
        }
    )
    await k
    return k


@pytest.fixture # type: ignore
def unknown_token_signer(
    base_url: str,
    unknown_keychain: Keychain
) -> AccessTokenSigner:
    return AccessTokenSigner(
        issuer=base_url,
        codec=PayloadCodec(
            signer=unknown_keychain,
            encrypter=unknown_keychain
        )
    )


"""Declares :class:`SettingsClientRepository`."""
from cbra.conf import settings # type: ignore
from cbra.utils import docstring
from .clientconfig import ClientConfig
from .types import IClient
from .types import IClientRepository


class SettingsClientRepository(IClientRepository):
    """A :class:`~cbra.ext.oauth2.types.IClientRepository` implementation
    that uses the :mod:`cbra` settings module as the source of the OAuth
    clients.
    
    The settings module must declare the ``OAUTH_CLIENTS`` attribute holding
    a dictionary that maps client identifiers (i.e. its keys) to dictionaries
    specifying the configuration for that specific client.
    """
    __module__: str = 'cbra.ext.oauth2'
    clients: dict[str, ClientConfig]
    model: type[ClientConfig] = ClientConfig

    def __init__(self) -> None:
        self.clients = {
            k: self.model.parse_obj({**v, 'client_id': k})
            for k, v in dict.items(getattr(settings, 'OAUTH_CLIENTS', {})) # type: ignore
        }

    @docstring(IClientRepository)
    async def exists(self, client_id: str) -> bool:
        return client_id in self.clients

    @docstring(IClientRepository)
    async def get(self, client_id: str) -> IClient:
        if client_id not in self.clients:
            raise self.ClientDoesNotExist
        return self.clients[client_id]


"""Declares :class:`RFC9068Token`."""
from typing import Any

from ckms.jose import PayloadCodec
from ckms.types import AudienceType
from ckms.types import ClaimSet

from .spaceseparatedlist import SpaceSeparatedList


class RFC9068Token(ClaimSet):
    iss: str
    aud: AudienceType
    exp: int
    sub: int | str
    client_id: str
    iat: int
    jti: str
    auth_time: int | None = None
    acr: str | None = None
    amr: list[str] | None = []
    scope: SpaceSeparatedList = SpaceSeparatedList()

    async def sign(self, codec: PayloadCodec, **kwargs: Any) -> str:
        """Sign the token using the given encoder."""
        return await codec.encode(
            payload={
                **self.dict(),
                'aud': list(self.aud)[0] if len(self.aud) == 1 else list(sorted(self.aud)),
                'scope': str(self.scope)
            },
            **kwargs
        )

    class Config(ClaimSet.Config):
        json_encoders = {
            **ClaimSet.Config.json_encoders,
            SpaceSeparatedList: lambda v: str.join(' ', sorted(v)) # type: ignore
        }


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .icredential import ICredential


class ClientCredential(ICredential):
    """The base class for OAuth 2.0 client credentials. A client credential
    obtains an access token credential from the authorization server.
    """
    __module__: str = 'cbra.ext.api.credential'
    client_id: str
    client_assertion_type: str

    def __init__(
        self,
        client_id: str
    ):
        self.client_id = client_id


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`MoneybirdWebhook`."""
from typing import TypeAlias

from cbra.ext import webhook
from .models import EventType


class MoneybirdWebhook(webhook.WebhookEndpoint):
    __module__: str = 'cbra.ext.moneybird'
    summary: str = "Moneybird"
    model: TypeAlias = EventType
    description: str = (
        "Receive a Moneybird event and queue it for upstream "
        "processing.\n\n"
        "Refer to https://developer.moneybird.com for additional "
        "documentation on the Moneybird API."
    )


# pylint: skip-file
from .dependency import Dependency
from .dependencylist import DependencyList
from .filedependencyspec import FileDependencySpec
from .symboldependencyspec import SymbolDependencySpec


__all__ = [
    'Dependency',
    'DependencyList',
    'FileDependencySpec',
    'SymbolDependencySpec',
]



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Iterable

import pydantic

from .authorizedscope import AuthorizedScope
from .types import AccessType
from .types import AuthorizationIdentifier
from .types import IAuthorization
from .types import RefreshToken
from .types import TokenException


IGNORED_SCOPES: set[str] = {"openid"}


class Authorization(pydantic.BaseModel, IAuthorization):
    """Represents to permission given to a Client by a Resource Owner."""
    id: int | None = pydantic.Field(
        default=None
    )

    version: int = pydantic.Field(
        default=0
    )

    client_id: str = pydantic.Field(
        default=...
    )

    sub: int | str = pydantic.Field(
        default=...
    )

    access_type: AccessType = pydantic.Field(
        default=AccessType.online
    )

    created: datetime.datetime = pydantic.Field(
        default_factory=datetime.datetime.utcnow
    )

    scope: dict[str, AuthorizedScope] = pydantic.Field(
        default={}
    )

    @classmethod
    def fromkey(
        cls,
        key: AuthorizationIdentifier
    ) -> 'Authorization':
        """Create a new :class:`Authorization` using the given key."""
        return cls(
            client_id=key.client_id,
            sub=key.sub
        )

    @property
    def key(self) -> AuthorizationIdentifier:
        return AuthorizationIdentifier(
            client_id=self.client_id,
            sub=self.sub
        )

    def allows_refresh(self) -> bool:
        """Return a boolean indicating if the :class:`Authorization` allows
        the issue of a refresh token.
        """
        return self.access_type == AccessType.offline

    def authorize(
        self,
        scope: str,
        now: datetime.datetime,
        expires: datetime.datetime | None = None
    ) -> None:
        """Authorizes the given scope `scope` for the Resource Owner
        and Client.
        """
        if self.is_authorized(scope) or scope in IGNORED_SCOPES:
            return
        self.scope[scope] = AuthorizedScope(
            name=scope,
            authorized=now,
            expires=expires
        )

    def create_refresh_token(
        self,
        scope: Iterable[str],
        include_granted: bool = False
    ) -> RefreshToken:
        """Create a *new* refresh token for the authorization."""
        assert self.id is not None # nosec
        if not self.allows_refresh():
            raise TokenException(
                code='invalid_grant',
                description=(
                    "The issue of refresh tokens is disabled by the Client or "
                    "was not authorized by the Resource Owner."
                )
            )
        if not self.is_authorized(scope):
            raise TokenException(
                code='invalid_scope',
                description=(
                    "The requested scope exceeds the scope that was granted by "
                    "the Resource Owner."
                )
            )
        scope = set(scope)
        if include_granted:
            scope |= self.get_active_scope()
        return RefreshToken.new(
            authorization_id=self.id,
            client_id=self.client_id,
            sub=self.sub,
            scope=list(sorted(scope))
        )

    def enable_refresh(self) -> None:
        """Enables the issue of refresh tokens."""
        self.access_type = AccessType.offline

    def expire_scope(self, now: datetime.datetime | None = None) -> None:
        """Expire authorized scopes that have an expiration date."""
        for scope in list(self.scope.values()):
            if not scope.is_expired(now):
                continue
            self.scope.pop(scope.name)

    def get_active_scope(self, now: datetime.datetime | None = None) -> set[str]:
        """Return the set of scopes that are currently active."""
        now = now or datetime.datetime.utcnow()
        return {
            scope.name
            for scope in self.scope.values()
            if not scope.is_expired()
        } | {"openid"}

    def is_authorized(self, scope: str | Iterable[str]) -> bool:
        """Return a boolean indicating if authorization was granted previously
        to the given scope.
        """
        if isinstance(scope, str):
            scope = {scope}
        return set(scope) <= self.get_active_scope()


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Iterable

from .iclient import IClient
from .isubject import ISubject


class IRefreshToken:
    __module__: str = 'cbra.ext.oauth2.types'

    def allows_scope(self, scope: str | Iterable[str] | None) -> bool:
        """Return a boolean indicating if the Refresh Token allows the given
        scope.
        """
        raise NotImplementedError

    async def generate(
        self,
        client: IClient,
        subject: ISubject,
        rotate: bool = False
    ) -> str | None:
        """Generate a new refresh token for the given `client`."""
        raise NotImplementedError

    async def verify(self, token: str) -> bool:
        """Return a boolean indicating if the given refresh token has a valid
        signature.
        """
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`GoogleTransport`."""
import aorta


class GoogleTransport(aorta.transport.GoogleTransport):
    command_project: str
    command_topic: str
    events_project: str
    events_topic: str

    def __init__(
        self,
        project: str,
        command_topic: str,
        events_topic: str,
        command_project: str | None = None,
        events_project: str | None = None
    ):
        super().__init__(project=project, topic_path=None)
        self.command_project = command_project or project
        self.command_topic = command_topic
        self.events_project = events_project or project
        self.events_topic = events_topic

    def get_topics(self, message: aorta.models.Message):
        topics = [self.events_topic, f'{self.events_topic}.{message.kind}']
        project = self.events_project
        if message.is_command():
            project = self.command_project
            topics = [self.command_topic]
        return [self.client.topic_path(project, x) for x in topics]



"""Declares :class:`ClientAssertionType`."""
import enum


class ClientAssertionType(str, enum.Enum):
    __module__: str = 'cbra.ext.oauth2.types'
    jwt_bearer = 'urn:ietf:params:oauth:client-assertion-type:jwt-bearer'


"""Declares :class:`QueryRunner`."""
import functools
from typing import Any


class QueryRunner:
    """The base class for all query runners."""
    __module__: str = 'cbra.query'

    async def connect(self):
        """Setup a connection to the persistent storage system or otherwise
        configure the :class:`QueryRunner` to begin executing queries.
        """
        pass

    async def close(self):
        """Close the connection(s) with the persistent storage system(s) and
        release all resources.
        """
        pass

    @functools.singledispatchmethod
    async def execute(self, query: Any) -> Any:
        raise NotImplementedError

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, *args, **kwargs):
        # Since the QueryRunner never does mutations, no special exception
        # handling is needed on exit.
        await self.close()



"""Declares :class:`EnvelopeMetadata`."""
import typing

import pydantic


class EnvelopeMetadata(pydantic.BaseModel):
    __module__: str = 'cbra.models'
    id: typing.Optional[typing.Union[int, str]] = pydantic.Field(
        default=None,
        title="Object identifier",
        description=(
            "Identifies the resource uniquely accross all resource versions. This "
            "property must not be set on `create` actions, unless otherwise specified."
        )
    )
    links: typing.Dict[str, str] = pydantic.Field(
        default={},
        title="Links",
        description=(
            "Additional links related to the resource. Read only."
        )
    )

    labels: dict[str, str] = pydantic.Field(
        default={},
        title="Labels",
        description=(
            "Map of string keys and values that can be used "
            "to organize and categorize (scope and select) objects."
        )
    )

    annotations: dict[str, typing.Any] = pydantic.Field(
        default={},
        title="Annotations",
        description=(
            "Annotations is an unstructured key value map stored "
            "with a resource that may be set by external tools to "
            "store and retrieve arbitrary metadata. Consult the "
            "documentation to determine if this resource supports "
            "annotations."
        )
    )


"""Declares :class:`RequestQueryMetaclass`."""
import typing

import pydantic

from .namedqueryargs import NamedQueryArgs
from .namedquerymetaclass import NamedQueryMetaclass


class RequestQueryMetaclass(NamedQueryMetaclass):

    def create_model(
        cls,
        name: str,
        bases: typing.List[type],
        hints: dict,
        fields: dict
    ) -> NamedQueryArgs:
        """Construct the model used to validate the query parameters."""
        return type(
            f'Request{name}',
            (pydantic.BaseModel,),
            {
                'limit': 100,
                'offset': 0,
                '__annotations__': {
                    'args' :super().create_model(cls, name, bases, hints, fields),
                    'limit': int,
                    'offset': int
                }
            }
        )



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`WebhookEndpoint`."""
import base64
import binascii
import urllib.parse
from typing import NoReturn
from typing import TypeAlias

from cbra.exceptions import WebhookMessageRejected
from cbra.ext import webhook
from cbra.ext.picqer.models import EventType
from cbra.parsers import JSONParser
from cbra.parsers import FormParser
from cbra.types import IParser
from .const import HEADER_DELIVERY_ID
from .const import HEADER_EVENT
from .const import HEADER_RESOURCE
from .const import HEADER_SIGNATURE
from .const import HEADER_SOURCE
from .const import HEADER_TOPIC
from .const import HEADER_WEBHOOK_ID
from .models import *


class WebhookEndpoint(webhook.WebhookEndpoint):
    """A :class:`cbra.ext.webhook.WebhookEndpoint` implementation
    that recognized incoming webhooks from WooCommerce (RESt API v3).
    """
    __module__: str = 'cbra.ext.woocommerce'
    allowed_events: set[str] = {"created", "updated", "deleted", "restored"}
    description: str = (
        "Receive a WooCommerce event and queue it for upstream "
        "processing.\n\n"
        "Refer to https://woocommerce.github.io/woocommerce-rest-api-docs "
        "for additional documentation on the WooCommerce v3 API."
    )
    echo: bool = False
    model: TypeAlias = EventType
    parsers: list[type[IParser]] = [
        JSONParser,
        FormParser
    ]
    summary: str = "WooCommerce (WP REST API integration v3)"
    wc_delivery_id: str | None
    wc_event: str | None
    wc_resource: str | None
    wc_signature: bytes | None
    wc_source: str | None
    wc_topic: str | None
    wc_webhook_id: str | None
    source: str | None = None

    def __init__(
        self,
        delivery_id: str | None = HEADER_DELIVERY_ID,
        event: str | None = HEADER_EVENT,
        resource: str | None = HEADER_RESOURCE,
        signature: str | None = HEADER_SIGNATURE,
        source: str | None = HEADER_SOURCE,
        topic: str | None = HEADER_TOPIC,
        webhook_id: str | None = HEADER_WEBHOOK_ID
    ):
        self.wc_delivery_id = delivery_id
        self.wc_event = event
        self.wc_resource = resource
        self.wc_source = source
        self.wc_signature = None
        self.wc_topic = topic
        self.wc_webhook_id = webhook_id
        if signature is not None:
            try:
                self.wc_signature = base64.b64decode(signature)
            except binascii.Error:
                raise WebhookMessageRejected(
                    message="The signature is malformed.",
                    detail=(
                        "The signature enclosed in the X-WC-Webhook-Signature "
                        "request header could not be interpreted as Base64."
                    )
                )
        if self.wc_source is not None:
            p = urllib.parse.urlparse(self.wc_source)
            self.source = p.netloc

    async def authenticate( # type: ignore
        self,
        dto: EventType
    ) -> NoReturn | None:
        """Verifies that the `X-WC-Signature` header was created using the
        preshared key.
        """
        if isinstance(dto, Ping):
            return
        await self.verify_request(
            signature=self.wc_signature or b'',
            secret=self.get_hmac_secret()
        )


"""Declares :class:`InjectedMessageHandler`."""
from typing import Any

import aorta


# These classes exist for legacy compatibility; cbra >-0.67.0 is
# fully compatible with Aorta.
class CommandHandler(aorta.CommandHandler):
    __abstract__: bool = True
    __module__: str = 'cbra'

    async def handle(self, command: Any) -> None: # type: ignore
        raise NotImplementedError


class EventListener(aorta.EventListener):
    __abstract__: bool = True
    __module__: str = 'cbra'

    async def handle(self, event: Any) -> None: # type: ignore
        raise NotImplementedError


# pylint: skip-file
import typing
from typing import Any
from typing import Callable

from .dependency import Dependency
from .environment import Environment
from .inheritdependencydecorator import InheritDependenciesDecorator
from .injected import Injected
from .instance import Instance
from .parameterless import Parameterless
from .provider import Provider
from .provider import _default # type: ignore


__all__ = [
    'clone',
    'is_satisfied',
    'provide',
    'resolve',
    'Dependency',
    'Environment',
    'Injected',
    'Instance',
    'Parameterless',
    'Provider'
]


def clone(*args: Any, **kwargs: Any) -> Callable[..., Any]:
    """Injects the dependencies of the given callable."""
    return InheritDependenciesDecorator(*args, **kwargs)


def environment(name: str) -> Any:
    return Environment(name)

def instance(dependency: type[Any] | str) -> Any:
    return Instance(dependency)


def inject(dependency: typing.Union[str, type]) -> typing.Any:
    return Injected(dependency)


def is_satisfied(name: str, using: Provider | None = None) -> bool:
    """Return a boolean indicating if the dependency is satisfied."""
    return (using or _default).is_satisfied(name)


def provide(
    name: str,
    value: object,
    force: bool = False,
    using: Provider | None = None
) -> None:
    """Register Python object `value` as a dependency under the key `name` and
    return the object.

    The `force` argument indicates if any existing dependency under `name`
    must be overwrriten. If `force` is ``False``, an exception is raised if
    `name` is already provided.
    """
    return (using or _default).provide(name, value, force)


def resolve(name: str, using: Provider | None = None) -> typing.Any:
    """Resolve a priorly injected dependency by its name."""
    return (using or _default).resolve(name)


async def on_boot(*args: Any, using: Provider | None = None, **kwargs: Any) -> None:
    """Provides a hook to execute logic when booting an application."""
    await (using or _default).on_boot(*args, **kwargs)


async def on_shutdown(*args: Any, using: Provider | None = None, **kwargs: Any) -> None:
    """Provides a hook to execute logic when shutting down an application."""
    await (using or _default).on_shutdown(*args, **kwargs)


"""Declares :class:`BaseSession`."""
from typing import Any
from typing import Awaitable

import fastapi
from ckms.jose import PayloadCodec


class BaseSession:
    awaiting: object = object()
    claims: dict[str, Any] | object = awaiting
    created: bool = False
    codec: PayloadCodec
    dirty: set[str]
    path: str= '/'
    request: fastapi.Request
    reserved_keys: set[str] = {"iat", "mod"}

    @classmethod
    def as_dependency(cls) -> Any:
        return fastapi.Depends(cls)

    @classmethod
    def configure(cls, **attrs: Any) -> Any:
        return type(cls.__name__, (cls,), attrs)

    async def add_to_response(self, response: fastapi.Response) -> None:
        raise NotImplementedError

    def clear(self) -> Awaitable[None] | None:
        raise NotImplementedError

    def get(self, key: str) -> Awaitable[Any] | Any:
        raise NotImplementedError

    def set(self, key: str, value: Any) -> Awaitable[None] | None:
        raise NotImplementedError

    def pop(self, key: str) -> Awaitable[None] | None:
        raise NotImplementedError


# type: ignore
"""An OAuth 2.x client that automatically discovers the servers'
capabilities.
"""
import asyncio
import collections
import datetime
import secrets
from typing import Any

import httpx
import jwcrypto.jwk
import jwcrypto.jwt

from cbra.ext.oauth2.ref.keys import KEY_BOB


OAUTH2_SERVER = "https://localhost:8000"


class SubjectClient(httpx.AsyncClient):
    client_id: str
    default_assertion_ttl: int = 30
    metadata_url: str = "/.well-known/oauth-authorization-server"
    private_key: jwcrypto.jwk.JWK
    subject: str
    _metadata: dict[str, Any] = collections.defaultdict(dict)

    @property
    def jwks_uri(self) -> str | None:
        """The URL of the authorization servers' JWKS."""
        return self._metadata[self.server].get('jwks_uri')

    @property
    def metadata(self) -> dict[str, Any]:
        """The metadata describing the server, as a dictionary."""
        return self._metadata[self.server]

    @property
    def token_endpoint(self) -> str | None:
        """The URL of the authorization servers' token endpoint."""
        return self._metadata[self.server].get('token_endpoint')

    def __init__(
        self,
        server: str,
        client_id: str,
        subject: str,
        private_key: bytes,
        **kwargs: Any
    ):
        super().__init__( # type: ignore
            base_url=server,
            **kwargs
        )
        self.client_id = client_id
        self.server = server
        self.private_key = jwcrypto.jwk.JWK.from_pem(private_key, None) # type: ignore
        self.subject = subject
        self._metadata = SubjectClient._metadata

    def create_assertion(self) -> str:
        """Creates an assertion to retrieve a token from the
        endpoint.
        """
        now = int(datetime.datetime.utcnow().timestamp())
        claims = {
            'jti': secrets.token_urlsafe(24),
            'iss': self.subject,
            'aud': self.token_endpoint,
            'sub': self.subject,
            'iat': now,
            'exp': now + self.default_assertion_ttl,
            'nbf': now,
        }
        token = jwcrypto.jwt.JWT(
            header={'alg': 'EdDSA'},
            claims=claims
        )
        token.make_signed_token(self.private_key)
        return token.serialize()

    async def get_access_token(self, scope: set[str]):
        response = await self.post(
            url=self.token_endpoint,
            json={
                'grant_type': "urn:ietf:params:oauth:grant-type:jwt-bearer",
                'client_id': self.client_id,
                'assertion': self.create_assertion(),
                'scope': str.join(' ', sorted(scope))
            }
        )
        if response.status_code != 200:
            raise Exception(response.text)
        dto = response.json()
        return dto['access_token']

    async def get_server_metadata(self) -> dict[str, Any]:
        """Return a dictionary containing the metadata that was discovered
        from the authorization server.
        """
        response = await self.get(self.metadata_url)
        response.raise_for_status()
        return response.json()

    async def __aenter__(self):
        client = await super().__aenter__()
        if self.server not in self._metadata:
            self._metadata[self.server] = await self.get_server_metadata()
        return self


async def main():
    params = {
        'client_id': 'jwt',
        'server': OAUTH2_SERVER,
        'verify': False,
        'subject': "bob@example.unimatrixone.io",
        'private_key': open(KEY_BOB, 'rb').read()
    }
    async with SubjectClient(**params) as client:
        print(client.token_endpoint)
        print(client.jwks_uri)
        print(client.create_assertion())
        print(await client.get_access_token({"openid", "read"}))


if __name__ == '__main__':
    asyncio.run(main())



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any

import httpx

from .credential import Credential


class Resource:
    """Provides an interface to interace with a resource server."""
    __module__: str = 'cbra.ext.service'
    credential: Credential
    http: httpx.AsyncClient
    logger: logging.Logger = logging.getLogger("uvicorn")
    server: str

    def __init__(
        self,
        *,
        server: str,
        credential: Credential
    ):
        if str.endswith(server, '/'):
            raise ValueError("The `server` parameter must not end with a slash.")
        self.credential = credential
        self.http = httpx.AsyncClient(base_url=server)
        self.server = server

    async def connect(self) -> None:
        """Setup the HTTP client for the specified resource server."""
        await self.http.__aenter__()
        self.logger.info("Connected to resource server %s", self.server)

    async def close(self) -> None:
        """Close the HTTP connection to the resource server."""
        await self.http.__aexit__(None, None, None)
        self.logger.info("Closed connection to resource server %s", self.server)

    async def request(
        self,
        *,
        method: str,
        path: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any
    ) -> httpx.Response:
        if not str.startswith(path, '/'):
            raise ValueError("Path must start with a slash.")
        response = await self.http.request( # type: ignore
            method=method,
            url=path,
            headers={
                'Authorization': f'Bearer {self.credential}',
                **(headers or {})
            },
            **kwargs
        )
        return response



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .types import IClient
from .types import IClientRepository


class NullClientRepository(IClientRepository):
    """A :class:`IClientRepository` implementation that has no clients."""

    async def exists(self, client_id: str) -> bool:
        return False

    async def get(self, client_id: str) -> IClient:
        raise self.ClientDoesNotExist


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import httpx
from ckms.utils import current_timestamp

from .icredential import ICredential


class AccessTokenCredential(ICredential):
    """An OAuth 2.0 access token."""
    __module__: str = 'cbra.ext.api.credential'
    expires: int
    token: str
    refresh_token: str | None

    def __init__(self, token: str, expires_in: int, refresh_token: str | None = None):
        self.token = token
        self.expires = current_timestamp() + expires_in
        self.refresh_token = refresh_token

    async def add_to_request(self, request: httpx.Request) -> None:
        request.headers['Authorization'] = f'Bearer {self.token}'


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .tokenexception import TokenException


class InvalidRefreshToken(TokenException):

    def __init__(self):
        super().__init__(
            code='invalid_grant',
            description="The refresh token is malformed."
        )


# pylint: skip-file
from typing import Any
from typing import Coroutine

import pydantic
from ckms.core.models import JSONWebSignature
from ckms.types import JSONWebKeySet
from ckms.types import JSONWebKey

from cbra.ext.oauth2.types import BaseSubject


class Subject(BaseSubject, pydantic.BaseModel):
    sub: str
    keys: list[JSONWebKey] = []
    client_id: str | None

    @property
    def jwks(self) -> JSONWebKeySet:
        return JSONWebKeySet(keys=self.keys)

    def allows_scope(self, scope: set[str]) -> bool:
        assert isinstance(scope, set)
        if self.client_id is None:
            return False
        return scope <= self.client_scope[self.client_id]

    def register_public_key(self, key: JSONWebKey) -> None:
        self.keys.append(key)

    def verify(
        self,
        jws: JSONWebSignature
    ) -> bool | Coroutine[Any, Any, bool]:
        return jws.verify(self.jwks, require_kid=False)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from ckms.core import get_default_keychain
from ckms.core import Keychain
from ckms.core import KeySpecification
from ckms.utils import b64encode_str

from .encryptedvalue import EncryptedValue


class PersistenceSecurity:
    """Provides an interface to encrypt/decrypt the data persistence layer
    (data-at-rest).
    """
    __module__: str = 'cbra.ext.service'
    keychain: Keychain = get_default_keychain()

    @property
    def encryption_key(self) -> KeySpecification:
        return self.keychain.get('pii')

    async def decrypt(self, ct: EncryptedValue) -> bytes | str:
        """Decrypt `ct`."""
        pt = await ct.decrypt(self.encryption_key)
        return pt

    async def encrypt(
        self,
        value: bytes | str,
        encoding: str = 'utf-8'
    ) -> EncryptedValue:
        """Encrypt `value."""
        applied_encoding = None
        if not isinstance(value, bytes):
            value = str.encode(value, encoding=encoding)
            applied_encoding = encoding
        ct = await self.encryption_key.encrypt(value)
        return EncryptedValue.parse_obj({
            'content': bytes(ct),
            'aad': ct.aad,
            'iv': ct.iv,
            'tag': ct.tag,
            'encoding': applied_encoding 
        })

    async def index(
        self,
        value: bytes | str,
        encoding: str = 'utf-8',
        key: KeySpecification | None = None
    ) -> str:
        """Create a secure index of `value`.
        
        If `value` is of type :class:`str`, use `encoding` as the character
        encoding when converting the string to bytes.
        """
        key = key or self.keychain.get('idx')
        if isinstance(value, str):
            value = str.encode(value, encoding=encoding)
        return b64encode_str(await key.sign(value))


# pylint: skip-file
# type: ignore
import functools
import typing

from .basefiltermodel import BaseFilterModel
from .filteroption import FilterOption
from .iresource import IResource
from .publicresource import PublicResource
from .resource import Resource


__all__ = [
    'action',
    'BaseFilterModel',
    'FilterOption',
    'IResource',
    'PublicResource',
    'Resource'
]


def action(
    name: typing.Optional[str] = None,
    detail: bool = False,
    method: str = 'GET',
    path: typing.Optional[str] = None
) -> typing.Callable[..., typing.Any]:
    def action_factory(func):
        func.action = dict(
            name=name or func.__name__,
            method=method,
            is_detail=detail,
            subpath=path
        )
        return func
    return action_factory


def response(
    status_code: int = 200,
    summary: typing.Optional[str] = None
) -> typing.Any:
    """Decorate an action handler with the given properties."""
    pass


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import NoReturn

from cbra.exceptions import HTTPException
from .promptexception import PromptException
from .redirecturl import RedirectURL


class AuthorizationException(HTTPException):
    """Base class for errors during an Authorization Request, not including
    configuration or request parameter errors."""
    __module__: str = 'cbra.ext.oauth2.types'
    log_level: str = 'INFO'

    def log(self, logger: logging.Logger) -> None:
        """Create a log record of this error."""
        raise NotImplementedError

    def raise_for_user(self, url: str, redirect_uri: RedirectURL) -> NoReturn:
        """Raises the exception with an indication that it must be
        displayed to the end-user, within the trust boundaries of
        the authorization server.
        """
        raise PromptException(url, {
            'return_uri': redirect_uri.error(error='access_denied')
        })


"""Declares :class:`DebugPrincipal`."""
from fastapi import Cookie

from cbra.types import IPrincipal
from cbra.utils import docstring


class DebugPrincipal(IPrincipal):
    """An :class:`~cbra.types.IPrincipal` implementation that retrieves
    the subject identifier from a cookie value. Used for testing purposes
    only.

    This principal implementation authenticates a request by looking up
    a cookie (specified by :attr:`cookie_name`) from the request and takes
    the content of the cookie as the subject identifier.
    """
    __module__: str = 'cbra.auth'
    cookie_name: str = 'subject'
    subject_identifier_type: type[int] | type[str] = str

    def __init__(self, sub: str | None = Cookie(None, alias=cookie_name)):
        self.sub = sub or ''

    @docstring(IPrincipal)
    def is_authenticated(self) -> bool:
        return bool(self)

    def __bool__(self) -> bool:
        return self.sub is not None


"""Declares :class:`IResource`."""
import typing

import fastapi.params

from cbra.types import IEndpoint
from .pathparameter import PathParameter


class IResource(IEndpoint):
    __module__: str = 'cbra.resource'
    document: bool = True
    filter_actions: list[str] = ['list']
    filter_options: list[fastapi.params.Query] = []
    is_detail: bool
    method: str
    model: typing.Any = None
    name: str
    name_article: str
    parent: type['IResource'] | None = None
    pluralname: str
    path_parameter: typing.Union[str, PathParameter]
    path_parameter_class: typing.Type[typing.Any]
    require_authentication: bool = True
    subresources: list[type['IResource']] = []
    verbose_name: str
    verbose_name_plural: str


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic


class OIDCClaimRequest(pydantic.BaseModel):
    essential: bool = False
    value: Any = None
    values: list[str | int] = pydantic.Field(
        default=[],
        max_items=10
    )

    @pydantic.validator('values')
    def validate_values(
        cls,
        values: list[Any]
    ) -> list[Any]:
        """Validates that the members of the input do not have excessive
        size.
        """
        for value in values:
            if isinstance(value, int):
                continue
            if isinstance(value, str) and len(value) > 512:
                raise ValueError("Requested claim value exceeds 512 characters.")
        return values

    def accepts(self, value: str) -> bool:
        """Return a boolean indicating if the :class:`OIDCClaimRequest`
        accepts the given value.
        """
        if (self.value is None) and not self.values:
            return True
        return value in  (self.values or [self.value])


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import dataclasses


@dataclasses.dataclass
class AuthorizationIdentifier:
    """Identifies the :class:`~cbra.ext.oauth2.types.Authorization`
    given to a specific Client by the Resource Owner.
    """
    client_id: str
    sub: int | str


# pylint: skip-file
from typing import TypeAlias

from .anonymousreadcorspolicy import AnonymousReadCorsPolicy
from .basecorspolicy import BaseCorsPolicy
from .defaultcorspolicy import DefaultPolicy
from .endpointcorspolicy import EndpointCorsPolicy
from .null import NullCorsPolicy


__all__ = [
    'AnonymousReadCorsPolicy',
    'BaseCorsPolicy',
    'CorsPolicyType',
    'DefaultPolicy',
    'EndpointCorsPolicy',
    'NullCorsPolicy'
]

CorsPolicyType: TypeAlias = BaseCorsPolicy


"""Create a :rfc:`9068` access token for use with the examples."""
import asyncio
import datetime
import secrets

import ckms
import ckms.jose


async def create_access_token(
    issuer: str = "https://localhost:8000",
    audience: str = "https://localhost:8000",
    ttl: int = 600,
) -> str:
    keychain = ckms.Keychain()
    keychain.register({
        'sig' :{
            'provider': 'local',
            'path': 'pki/ed448.key',
            'use': 'sig'
        }
    })
    await keychain.setup()
    now = int(datetime.datetime.utcnow().timestamp())
    codec = ckms.jose.PayloadCodec(signer=keychain)
    jwt = ckms.jose.JSONWebToken(
        client_id="test",
        jti=secrets.token_hex(),
        iss=issuer,
        aud=audience,
        sub="bob@cbra.unimatrixapis.localhost",
        iat=now,
        exp=now+ttl,
        nbf=now - 10,
        scope="god"
    )
    meta = keychain.get_key_metadata('sig')
    jws = await codec.encode(jwt)\
        .sign(
            algorithm='EdDSA',
            using='sig',
            claims={'kid': meta.get_signing_kid()}
        )
    return bytes.decode(bytes(jws))


if __name__ == '__main__':
    result = asyncio.run(create_access_token())
    print(result)


"""Declares :class:`IRequestHandler`."""
import datetime
import inspect
import types
import typing
from typing import Any

import fastapi
import pydantic

from .icontentnegotiation import IContentNegotiation
from .icorspolicy import ICorsPolicy
from .iparser import IParser
from .iprincipal import IPrincipal
from .irenderer import IRenderer
from .iresponsedigest import IResponseDigest
from .nullquerymodel import NullQueryModel
from .queryoptions import QueryOptions
from .request import Request


class IRequestHandler:
    """The interface for objects that handle a request for a
    specific HTTP method.
    """
    RouterType: typing.Any = typing.Union[
        fastapi.FastAPI,
        fastapi.APIRouter
    ]
    body_methods: typing.Set[str] = {"POST", "PUT", "PATCH"}

    #: Specifies the Cross-Origin Resource Sharing (CORS)
    #: policy used for this endpoint.
    cors_policy: type[ICorsPolicy]

    #: The default digest algorithm if none was selected.
    default_digest: typing.Optional[typing.Type[IResponseDigest]] = None

    #: The default renderer if none was matched based on the content negotiation
    #: headers. If :attr:`default_renderer` is ``None``, then failure to negotiate
    #: a response content type leads to an error response.
    default_renderer: type[IRenderer] | None = None

    default_response_type: str
    default_response: typing.Dict[str, typing.Any]
    default_response_code: int = 200

    #: Description of the endpoint for the OpenAPI schema.
    description: str | None = None

    #: Indicates if there should be automatic documentation.
    document: bool = True

    #: The :class:`cbra.types.IResponseDigest` instance that was selected
    #: to calculate the digest for the response.
    digest: typing.Optional[IResponseDigest] = None

    #: The list of available digest algorithms.
    digest_algorithms: typing.List[typing.Type[IResponseDigest]] = []

    #: The HTTP request method that is handled by the request handler.
    method: str | None = None

    #: Supported HTTP methods by this endpoint implementation.

    #: The :class:`pydantic.BaseModel` implementation used to parse the
    #: request body, or a union thereof.
    model: type[pydantic.BaseModel] | types.UnionType | None = None

    #: The default mount path of the request handler. Must not start with
    #: a slash.
    mount_path: str | None = None

    #: The name through which the request handler can be revered. If not
    #: name is specified, it is generated from the request handler.
    name: str | None = None

    #: The default implementation to negotiate content encoding, langauge
    #: and response digest.
    negotation: type[IContentNegotiation]

    #: The list of parser classes that known how to interpret the
    #: request body, based on the ``Content-Type`` header.
    parsers: list[type[IParser]] = []

    #: The parser that was selected for the current request.
    parser: IParser | None = None

    #: The path at which the request handler is mounted.
    path: str

    #: The authenticated principal of the request.
    principal: IPrincipal | None

    #: Factory function resolving the credentials provided with an
    #: HTTP request (if any, such as session or header authentication)
    #: to a :class:`IPrincipal` implementation.
    principal_factory: typing.Union[
        typing.Callable[..., typing.Optional[IPrincipal]],
        typing.Coroutine[typing.Any, typing.Any, typing.Optional[IPrincipal]],
        typing.Any
    ] = (lambda: None)

    #: Indicates if the endpoint is queryable.
    queryable: bool = False

    #: The current query provided through the request query
    #: parameters.
    query: NullQueryModel

    #: A :class:`fastapi.Request` instance that is being handled
    #: by the request handler.
    request: Request

    #: The list of renderers used to render response bodies.
    renderers: list[type[IRenderer]] = []

    #: The renderer that is current selected for the response.
    renderer: IRenderer | None = None

    #: Indicates if only authenticated requests are allowed.
    require_authentication: bool = False

    #: A mapping of response codes to OpenAPI response descriptions.
    responses: dict[int | str, typing.Any] = {}

    #: Description of the response used in the OpenAPI schema.
    response_description: str = "Success."

    #: The response class that is used by the request handler.
    response_class: type[fastapi.responses.Response] = fastapi.responses.JSONResponse

    #: The response model, if not inferred from the :meth:`handle()`
    #: signature.
    response_model: type[pydantic.BaseModel] | types.UnionType | None = None

    #: Indicates which fields should be excluded from the response
    #: model.
    response_model_exclude: set[str] | dict[str, Any] = set()


    #: Indicates if the response model must be serialized using the aliases
    #: defined on its :class:`pydantic.FieldInfo` members.
    response_model_by_alias: bool = False

    #: Indices if the defaults of the response model must be excluded when
    #: serializing.
    response_model_exclude_defaults: bool = False

    #: Indices if the ``None`` values of the response model must be excluded when
    #: serializing.
    response_model_exclude_none: bool = False

    #: The return type of the method handler (DEPRECATED).
    returns: type[pydantic.BaseModel] | types.UnionType | None = None

    #: The summary shown in the OpenAPI schema.
    summary: str | None = None

    #: OpenAPI tags for the endpoint.
    tags: list[str] = []

    @classmethod
    def add_to_router(
        cls,
        *,
        app: RouterType,
        base_path: str,
        **kwargs: typing.Any
    ) -> None:
        raise NotImplementedError

    @classmethod
    def get_path_signature(cls, detail: bool = False) -> list[inspect.Parameter]:
        return []

    @classmethod
    def get_responses(cls) -> dict[int | str, typing.Any]:
        """Return a mapping associating response codes to OpenAPI
        metadata.
        """
        raise NotImplementedError

    def is_authenticated(self) -> bool:
        """Return a boolean indicating if the request is authenticated."""
        raise NotImplementedError

    def setup(
        self,
        *,
        request: fastapi.Request,
        path: str,
        principal: typing.Optional[IPrincipal],
        parser: IParser,
        renderer: IRenderer,
        cors: ICorsPolicy,
        digest: typing.Optional[IResponseDigest],
        query: typing.Any,
        session: typing.Any
    ) -> None:
        self.cors = cors
        self.digest = digest
        self.parser = parser
        self.path = path
        self.principal = principal
        self.query = query
        self.renderer = renderer
        self.request = request
        self.session = session

    def serialize_response(
        self,
        request: fastapi.Request,
        context: typing.Dict[typing.Any, typing.Any]
    ) -> str:
        raise NotImplementedError

    async def on_exception(
        self,
        request: fastapi.Request,
        exception: Exception
    ) -> fastapi.Response:
        raise NotImplementedError



"""Declares :class:`TokenType`."""
import enum


class TokenType(str, enum.Enum):
    access_token = "urn:ietf:params:oauth:token-type:access_token"
    refresh_token = "urn:ietf:params:oauth:token-type:refresh_token"
    id_token = "urn:ietf:params:oauth:token-type:id_token"
    saml1 = "urn:ietf:params:oauth:token-type:saml1"
    saml2 = "urn:ietf:params:oauth:token-type:saml2"
    jwt = "urn:ietf:params:oauth:token-type:jwt"


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Any

import pydantic


def is_openapi_schema(
    value: Any
) -> bool:
    """Return a boolean indicating if the object may be used as an OpenAPI
    schema.
    """
    return is_pydantic_model(value)


def is_pydantic_model(value: Any) -> bool:
    """Return a boolean indicating if the value is a :mod:`pydantic`
    model.
    """
    return inspect.isclass(value) and issubclass(value, pydantic.BaseModel)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect

import pydantic
from ckms.core import KeySpecification
from ckms.types import CipherText
from ckms.utils import b64encode_str
from ckms.utils import b64decode


class EncryptedValue(pydantic.BaseModel):
    content: str
    iv: str | None = None
    aad: str | None = None
    tag: str | None = None
    encoding: str | None = None

    @property
    def ct(self) -> CipherText:
        return CipherText(
            buf=b64decode(self.content),
            iv=b64decode(self.iv or '') or None,
            aad=b64decode(self.aad or '') or None,
            tag=b64decode(self.tag or '') or None
        )

    @pydantic.validator('aad', pre=True)
    def preprocess_aad(cls, value: bytes | str | None) -> str | None:
        return cls.preprocess_bytes(value=value)

    @pydantic.validator('content', pre=True)
    def preprocess_content(cls, value: bytes | str | None) -> str | None:
        return cls.preprocess_bytes(value=value)

    @pydantic.validator('iv', pre=True)
    def preprocess_iv(cls, value: bytes | str | None) -> str | None:
        return cls.preprocess_bytes(value=value)

    @pydantic.validator('tag', pre=True)
    def preprocess_tag(cls, value: bytes | str | None) -> str | None:
        return cls.preprocess_bytes(value=value)

    @classmethod
    def preprocess_bytes(cls, value: bytes | str | None) -> str | None:
        if isinstance(value, bytes):
            value = b64encode_str(value)
        return value

    async def decrypt(self, key: KeySpecification) -> bytes | str:
        pt = key.decrypt(self.ct)
        if inspect.isawaitable(pt):
            pt = await pt
        assert isinstance(pt, bytes) # nosec
        if self.encoding:
            pt = bytes.decode(pt, self.encoding)
        return pt


"""Declares :class:`IContentNegotiation`."""
import typing
import warnings

import fastapi

from cbra.headers import ACCEPT
from cbra.headers import DIGEST_SCHEMA
from cbra.headers import ACCEPT_LANGUAGE
from cbra.headers import CONTENT_ENCODING
from cbra.headers import CONTENT_LENGTH
from cbra.headers import CONTENT_TYPE
from cbra.headers import WANTS_DIGEST
from .iparser import IParser
from .irenderer import IRenderer
from .iresponsedigest import IResponseDigest


class IContentNegotiation:
    accept: typing.Optional[str]
    default_response_encoding: str | None = None
    content_type: typing.Optional[str] = None
    content_encoding: typing.Optional[str] = None
    content_length: typing.Optional[int] = None
    wants_digest: typing.Optional[str] = None
    request: fastapi.Request

    @classmethod
    def get_response_headers(cls) -> dict[str, typing.Any]:
        return {'Digest': DIGEST_SCHEMA}

    @classmethod
    def has_request_body(cls) -> bool:
        raise NotImplementedError
 
    @classmethod
    def has_response_body(cls) -> bool:
        raise NotImplementedError


    def __init__(
        self,
        *,
        request: fastapi.Request,
        accept: typing.Optional[str] = ACCEPT,
        accept_language: str | None = ACCEPT_LANGUAGE,
        wants_digest: str | None = WANTS_DIGEST,
        content_encoding: str | None = CONTENT_ENCODING,
        content_length: int | None = CONTENT_LENGTH,
        content_type: str | None = CONTENT_TYPE,
    ):
        if wants_digest is not None:
            wants_digest = str.lower(wants_digest)
        self.request = request
        self.accept = accept or self.default_response_encoding
        self.accept_language = accept_language
        self.wants_digest = wants_digest
        self.content_encoding = content_encoding
        self.content_length = content_length
        self.content_type = content_type

    def select_digest(
        self,
        algorithms: typing.List[typing.Type[IResponseDigest]],
        default: typing.Optional[typing.Type[IResponseDigest]] = None
    ) -> typing.Optional[IResponseDigest]:
        raise NotImplementedError

    def select_parser(
        self,
        parsers: typing.List[typing.Type[IParser]]
    ) -> IParser:
        raise NotImplementedError

    def select_renderer(
        self,
        renderers: typing.List[typing.Type[IRenderer]],
        format_suffix: typing.Optional[str] = None,
        default: type[IRenderer] | None = None
    ) -> IRenderer:
        """Return the list of renderers in order of precedence."""
        raise NotImplementedError


# pylint: skip-file
import typing

import aorta
import fastapi
from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import JSONWebKeySet
from cbra.messagepublisher import MessagePublisher

from .types import IApplication


__all__ = [
    'ApplicationJWKS',
    'CurrentHost',
    'CurrentServer',
    'ServerCodec',
    'ServerKeychain',
    'ServerPublisher',
]


def get_client_host(request: fastapi.Request) -> str:
    host = '0.0.0.0'
    if request.client:
        host = request.client.host
    return host


def get_current_application(request: fastapi.Request) -> IApplication:
    return typing.cast(IApplication, request.app)


def get_current_host(request: fastapi.Request) -> str:
    return request.url.netloc


def get_current_server(request: fastapi.Request) -> str:
    return f'{request.url.scheme}://{request.url.netloc}'


def get_server_codec(request: fastapi.Request) -> PayloadCodec:
    return request.app.codec


def get_server_jwks(
    request: fastapi.Request
) -> JSONWebKeySet:
    return request.app.get_jwks()


def get_server_keychain(request: fastapi.Request) -> Keychain:
    return request.app.keychain


ClientHost: str = fastapi.Depends(get_client_host)

CurrentApplication: IApplication = fastapi.Depends(get_current_application)

CurrentHost: str = fastapi.Depends(get_current_host)

CurrentServer: str = fastapi.Depends(get_current_server)

ServerPublisher: aorta.MessagePublisher = fastapi.Depends(MessagePublisher)

ServerCodec: PayloadCodec = fastapi.Depends(get_server_codec)

ServerKeychain: Keychain = fastapi.Depends(get_server_keychain)

ApplicationJWKS: JSONWebKeySet = fastapi.Depends(get_server_jwks)



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied.
import pathlib

from ckms.types import JSONWebKey


KEY_DIR = pathlib.Path(__file__).parent
KEY_BOB = KEY_DIR.joinpath('bob.key')


alice: JSONWebKey = JSONWebKey.frompem(
    open(KEY_DIR.joinpath('alice.key'), 'rb').read(),
    kid='sig',
    alg='EdDSA',
    use='sig',
    key_ops=['sign']
)


bob: JSONWebKey = JSONWebKey.frompem(
    open(KEY_DIR.joinpath('bob.key'), 'rb').read(),
    kid='sig',
    alg='EdDSA',
    use='sig',
    key_ops=['sign']
)


trudy: JSONWebKey = JSONWebKey.frompem(
    open(KEY_DIR.joinpath('trudy.key'), 'rb').read(),
    kid='sig',
    alg='EdDSA',
    use='sig',
    key_ops=['sign']
)


"""Declares :class:`JWTBearerAssertionGrant`."""
import typing

import pydantic

from .basegrant import BaseGrant
from .granttype import GrantType
from .spaceseparatedlist import SpaceSeparatedList


class JWTBearerAssertionGrant(BaseGrant):
    __module__: str = 'cbra.ext.oauth2.types'

    grant_type: typing.Literal[GrantType.jwt_bearer] = pydantic.Field(
        default=...,
        title="Grant type",
        description=(
            "Must be `urn:ietf:params:oauth:grant-type:jwt-bearer`."
        ),
        example="urn:ietf:params:oauth:grant-type:jwt-bearer"
    )

    scope: SpaceSeparatedList | None = pydantic.Field(
        default=[],
        title="Scope",
        description=(
            "The requested scope that was previously granted to the subject "
            "or client."
        )
    )

    assertion: str = pydantic.Field(
        default=None,
        title="Assertion",
        description=(
            "The digitally signed document asserting the identity of the "
            "subject.\n\n"
            f"For grants of type `{GrantType.jwt_bearer.value}`, "
            "the `assertion` parameter contains a single JWT. The claims "
            "`exp`, `iat`, `nbf` must be present, in addition to the following"
            " claims:"
            "\n- `iss` - the OAuth 2.0 client id if the assertion is self-issued, "
            "otherwise the identifier of the third-party issuer."
            "\n- `aud` - the URL of the token endpoint."
            "\n- `sub` - identifies the subject on behalf of whom the access token "
            "is requested, i.e. the resource owner."
            "\n- `jti` - a randomly generated identifier for this specific assertion."
        )
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from ckms.types import IKeychain

from cbra import Application
from cbra.conf import settings
from .clientkeysendpoint import ClientKeysEndpoint
from .personaldatahandler import PersonalDataHandler
from .serviceclient import ServiceClient
from .serviceidentity import ServiceIdentity


class Service(Application):
    """An API server that integrates various :mod:`cbra` components."""
    __module__: str = 'cbra.ext.service'
    client_keychain: IKeychain
    identity: ServiceIdentity
    pii: PersonalDataHandler

    async def boot(self):
        self.client_keychain = self.keychain.tagged('oauth2-client')
        self.add(ClientKeysEndpoint.new(keychain=self.client_keychain))

        self.pii = PersonalDataHandler(self.keychain)
        self.client = ServiceClient(
            server=settings.OAUTH2_SERVER,
            identity=ServiceIdentity(
                client_id=settings.OAUTH2_SERVICE_CLIENT,
                keychain=self.client_keychain
            ),
            encrypter=self.pii
        )
        return await super().boot()

    async def teardown(self):
        await self.client.on_teardown()
        return await super().teardown()


# pylint: skip-file
import pytest

from ..symboldependencyspec import SymbolDependencySpec


@pytest.mark.asyncio
async def test_resolve_builtin():
    spec = SymbolDependencySpec(
        name='ExampleDependency',
        qualname='int'
    )
    assert await spec.resolve() == int


@pytest.mark.asyncio
async def test_resolve_symbol(symbol, package_name):
    spec = SymbolDependencySpec(
        name='ExampleDependency',
        qualname=f'{package_name}.conftest.SYMBOL'
    )
    assert await spec.resolve() == symbol



# Copyright (C) 2022 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from ckms.core import Keychain
from fastapi.security import HTTPBearer
from fastapi.security import HTTPAuthorizationCredentials

from cbra.ext.oauth2 import RFC9068Principal
from cbra.ext.oauth2.params import LocalIssuer
from cbra.params import ServerKeychain
from cbra.resource import Resource
from .accesstokenprincipal import AccessTokenPrincipal
from .nullprincipal import NullPrincipal


class ServiceResource(Resource):
    """A :class:`cbra.resource.Resource` implementation that authenticates
    a service with a :rfc:`9068` access token.
    """
    __abstract__: bool = True
    __module__: str = 'cbra.ext.service'
    principal: AccessTokenPrincipal
    security = HTTPBearer(auto_error=False)
    trusted_issuers: set[str] = set()

    @classmethod
    async def principal_factory(
        cls,
        request: fastapi.Request,
        issuer: str = LocalIssuer,
        bearer: HTTPAuthorizationCredentials | None = fastapi.Depends(security),
        keychain: Keychain = ServerKeychain
    ) -> AccessTokenPrincipal | NullPrincipal:
        f = RFC9068Principal(
            auto_error=cls.require_authentication,
            principal_factory=AccessTokenPrincipal.fromclaimset,
            trusted_issuers=cls.trusted_issuers
        )
        return await f(
            request=request,
            issuer=issuer,
            bearer=bearer,
            keychain=keychain
        ) or NullPrincipal()

    def is_authenticated(self) -> bool:
        """Return a boolean indicating if the request is authenticated."""
        return self.principal is not None and self.principal.is_authenticated()


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets

from ckms.jose import PayloadCodec
from ckms.types import IKeychain
from ckms.utils import current_timestamp


class ServiceIdentity:
    """Represents the identity of the service, using its signing and
    encryption keys.
    """
    __module__: str = 'cbra.ext.service'
    client_id: str
    codec: PayloadCodec
    keychain: IKeychain
    ttl: int = 60

    def __init__(self, client_id: str, keychain: IKeychain):
        self.client_id = client_id
        self.keychain = keychain
        self.codec = PayloadCodec(
            encrypter=self.keychain,
            decrypter=self.keychain,
            signer=self.keychain
        )

    async def get_client_assertion(
        self,
        token_endpoint: str
    ) -> str:
        """Encode a client assertion used to authenticate with the authorization
        server.
        """
        now = current_timestamp()
        return await self.codec.encode(
            payload={
                'aud': token_endpoint,
                'exp': now + self.ttl,
                'iat': now,
                'iss': self.client_id,
                'jti': secrets.token_urlsafe(48),
                'nbf': now,
                'sub': self.client_id,
            },
            signers=[ # type: ignore
                x for x in self.keychain
                if (not x.is_public() and x.can_sign())
            ]
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os

import httpx
from ckms.core import parse_dsn
from ckms.core import parse_spec
from ckms.core import KeySpecification
from ckms.types import ServerMetadata

from .accesstoken import AccessTokenCredential
from .icredential import ICredential
from .jwtprivatekeyclientcredential import JWTPrivateKeyClientCredential


class ApplicationClientCredential(ICredential):
    """A credential implementation that uses an application-specific
    signing/encryption keypair to retrieve an access token from an
    OAuth 2.x authorization server.
    """
    access_tokens: dict[tuple[str, str], AccessTokenCredential] = {}
    server: str | None = os.getenv('OAUTH_SERVER')
    client_id: str | None = os.getenv('OAUTH_SERVICE_CLIENT')
    client_credential: JWTPrivateKeyClientCredential
    metadata: ServerMetadata | None = None
    encryption_key: KeySpecification | None = None
    signing_key: KeySpecification | None = None

    def __init__(self):
        if os.getenv('APP_ENCRYPTION_KEY'):
            self.encryption_key = parse_spec(parse_dsn(os.environ['APP_ENCRYPTION_KEY']))
        if os.getenv('APP_SIGNING_KEY'):
            self.signing_key = parse_spec(parse_dsn(os.environ['APP_SIGNING_KEY']))
        self.access_tokens = type(self).access_tokens

    async def add_to_request(self, request: httpx.Request) -> None:
        """Add the access token to a request as the ``Authorization``
        header.

        If no access token is maintained for the given audience, then invoke the
        token endpoint of the authorization server to obtain one.
        """
        assert self.client_id is not None # nosec
        _, audience = k = (
            self.client_id,
            f'{request.url.scheme}://{request.url.netloc.decode()}'
        )
        at = self.access_tokens.get(k)
        if at is None:
            at = self.access_tokens[k] = await self.obtain(audience=audience)
        return await at.add_to_request(request)

    async def obtain(self, audience: str) -> AccessTokenCredential:
        """Obtain an access token from the authorization server using the Client
        Credentials grant.
        """
        assert self.client_id is not None # nosec
        assert self.server is not None # nosec
        assert self.signing_key is not None # nosec
        await self.signing_key
        if self.encryption_key is not None:
            await self.encryption_key
        async with httpx.AsyncClient() as client:
            if self.metadata is None:
                self.metadata = await ServerMetadata.discover(
                    client=client,
                    issuer=self.server
                )
            assert self.metadata is not None # nosec
            endpoint = self.metadata.token_endpoint
        credential = JWTPrivateKeyClientCredential(
            client_id=self.client_id,
            token_endpoint=endpoint,
            signing_key=self.signing_key,
            encryption_key=None
        )
        return await credential.obtain(audience=audience)


# pylint: disable=W0611
"""This module is just a proxy to :mod:`unimatrix.conf`."""
from typing import cast
from typing import Any

from unimatrix.conf import settings # type: ignore


__all__: list[str] = ['settings']

settings: Any = cast(Any, settings)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`SessionGrant`"""
import typing

import pydantic

from .granttype import GrantType
from .scopedgrant import ScopedGrant


class SessionGrant(ScopedGrant):
    __module__: str = 'cbra.ext.oauth2.types'

    grant_type: typing.Literal[GrantType.session] = pydantic.Field(
        default=...,
        title="Grant type",
        description=(
            "Must be `session`."
        ),
        example="session"
    )

    session: str = pydantic.Field(
        default=...,
        title="Session",
        description=(
            "A JWT representing an authenticated session, signed by a "
            "key that is trusted by the authorization server, and "
            "optionally encrypted."
        )
    )


"""Declares :class:`ListQueryResult`."""
import typing

import pydantic


class ListQueryResult(pydantic.BaseModel):
    __module__: str = 'cbra.query'

    total: int
    offset: int
    limit: int = 100
    objects: typing.List[dict] = []



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from typing import Any

from ckms.jose import PayloadCodec
from ckms.utils import current_timestamp


class AccessTokenSigner:
    __module__: str = 'cbra.ext.oauth2'
    client_id: Any
    codec: PayloadCodec
    default_ttl: int
    issuer: str
    entropy: int
    signing_keys: list[str]

    def __init__(
        self,
        *,
        issuer: str,
        codec: PayloadCodec,
        client_id: Any = None,
        entropy: int = 48,
        default_ttl: int = 600,
        signing_keys: list[str] | None = None
    ):
        self.client_id = client_id
        self.codec = codec
        self.default_ttl = default_ttl
        self.entropy = entropy
        self.issuer = issuer
        self.signing_keys = signing_keys or []

    def initialize_claims(
        self,
        now: int | None = None,
        ttl: int | None = None,
        **claims: Any
    ) -> dict[str, int | str]:
        """Initialize the default claims ``exp``, ``iat``, ``jti``, ``nbf``
        and ``iss``.
        """
        now = now or current_timestamp()
        return {
            'exp': now + (ttl or self.default_ttl),
            'iat': now,
            'iss': self.issuer,
            'jti': secrets.token_urlsafe(48),
            'nbf': now
        }

    async def sign(
        self,
        *,
        audience: str,
        sub: int | str,
        client_id: Any = None,
        ttl: int | None = None,
        now: int | None = None,
        signing_keys: list[str] | None = None,
        scope: str | set[str] | list[str] | None = None,
        content_type: str = 'at+jwt',
        **claims: Any
    ) -> str:
        """Sign a :rfc:`9068` access token with the given claims."""
        if not isinstance(scope, str):
            scope = str.join(' ', list(sorted(scope or [])))
        if not (signing_keys or self.signing_keys):
            raise TypeError("The `signing_keys` parameter can not be `None`.")
        if not (client_id or self.client_id):
            raise TypeError("The `client_id` parameter can not be `None`.")
        return await self.codec.encode(
            payload={
                **self.initialize_claims(now=now, ttl=ttl),
                **claims,
                'aud': audience,
                'client_id': client_id or self.client_id,
                'sub': sub,
                'scope': scope
            },
            content_type=content_type,
            signing_keys=signing_keys or self.signing_keys
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any
from typing import Literal

import pydantic


class Event(pydantic.BaseModel):
    __module__: str = 'cbra.ext.picqer.models'
    idhook: int = pydantic.Field(
        default=...,
        title="Webhook ID",
        description=(
            "Identifies the webhook that sent the event."
        )
    )

    name: str = pydantic.Field(
        default=...,
        title="Name",
        description="Human-readable name of the webhook."
    )

    event: str

    event_triggered_at: datetime.datetime = pydantic.Field(
        default=...,
        title="Triggered at",
        description=(
            "The date/time at which the event was triggered."
        )
    )

    data: dict[str, Any] = pydantic.Field(
        default=...,
        title="Data",
        description=(
            "The object on which the event specified by the `event` "
            "attribute occurred."
        )
    )

    @classmethod
    def new(cls, *, name: str, event: str) -> type['Event']:
        """Create a new model for the given event `event`."""
        return type(name, (cls,), {
            'event': pydantic.Field(
                default=...,
                title="Event",
                description=f'Is always `{event}`.'
            ),
            'Config': type('Config', (cls.Config,), {
                'title': f'Picqer{name}'
            }),
            '__annotations__': {
                'event': Literal[event] # type: ignore
            }
        })


import cbra
import cbra.resource


class GrandChildResource(cbra.resource.PublicResource):
    path_parameter: str = 'grandchild_id'

    async def list(self, parent_id: int, child_id: int):
        return {'parent_id': parent_id, 'child_id': child_id}

    async def retrieve(
        self,
        parent_id: int,
        child_id: int,
        grandchild_id: int
    ):
        return {
            'parent_id': parent_id,
            'child_id': child_id,
            'grandchild_id': grandchild_id
        }


class ChildResource(cbra.resource.PublicResource):
    path_parameter: str = 'child_id:int'
    subresources = [GrandChildResource]

    async def list(self, parent_id: int):
        return {'parent_id': parent_id}

    async def retrieve(self, parent_id: int, child_id: int):
        return {'parent_id': parent_id, 'child_id': child_id}


class ParentResource(cbra.resource.PublicResource):
    subresources: list[type[cbra.resource.Resource]] = [
        ChildResource
    ]
    path_parameter: str = 'parent_id'

    async def list(self):
        pass


app = cbra.Application()
app.add(ParentResource)


if __name__ == '__main__':
    cbra.run('__main__:app', reload=True)


"""Declares :class:`APIRoute`."""
import typing

import fastapi
import fastapi.routing
import starlette.requests
import starlette.responses
import starlette.types
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint


__all__ = ['APIRoute']


class APIRoute(fastapi.routing.APIRoute):
    app: fastapi.FastAPI
    server: typing.Any
    middleware: list[Middleware]

    @classmethod
    def with_server(
        cls,
        server: fastapi.APIRouter
    ) -> type[fastapi.routing.APIRoute]:
        """Add the OAuth 2.0 router to the :class:`APIRoute` and set
        the appropriate middleware so that other components can detect
        the server.
        """
        return type(cls.__name__, (cls,), {
            'middleware': [
                Middleware(AuthorizationServerMiddleware, server=server)
            ]
        })

    def __init__(self, *args: typing.Any, **kwargs: typing.Any):
        super().__init__(*args, **kwargs)
        for cls, options in reversed(self.middleware or []): # type: ignore
            self.app = cls(self.app, **options)


class AuthorizationServerMiddleware(BaseHTTPMiddleware):
    server: fastapi.APIRouter

    def __init__(
        self,
        app: starlette.types.ASGIApp,
        server: fastapi.APIRouter
    ) -> None:
        super().__init__(app)
        self.server = server

    async def dispatch(
        self,
        request: starlette.requests.Request,
        call_next: RequestResponseEndpoint
    ) -> starlette.responses.Response:
        """Add the :class:`~cbra.ext.oauth2.AuthorizationServer` to the
        request object.
        """
        request.scope['oauth2'] = self.server
        return await call_next(request)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cbra.ext.oauth2.types import IPrincipal


class NullPrincipal(IPrincipal): # type: ignore

    def is_authenticated(self) -> bool:
        return False

    def has_scope(self, scope: str | set[str]) -> bool:
        return False


"""Declares :class:`IResponseDigest`."""


class IResponseDigest:
    __module__: str = 'cbra.types'
    algorithm_name: str

    def calculate(self, message: bytes) -> str:
        raise NotImplementedError


"""Declares :class:`ClientCredentialsGrant`"""
import typing

import pydantic

from .basegrant import BaseGrant
from .granttype import GrantType
from .spaceseparatedlist import SpaceSeparatedList


class ClientCredentialsGrant(BaseGrant):
    __module__: str = 'cbra.ext.oauth2.types'

    grant_type: typing.Literal[GrantType.client_credentials] = pydantic.Field(
        default=...,
        title="Grant type",
        description=(
            "Must be `client_credentials`."
        ),
        example="client_credentials"
    )

    scope: SpaceSeparatedList | None = pydantic.Field(
        default=[],
        title="Scope",
        description=(
            "The requested scope that the client wants to obtain an access "
            "token for."
        )
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from cbra.ext.oauth2.types import IPrincipal
from cbra.ext.oauth2.types import RFC9068Token


class ClientPrincipal(IPrincipal, RFC9068Token): # type: ignore
    sub: int | str

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        if values.get('sub') != values.get('client_id'):
            raise ValueError(
                "Token was not granted trough the "
                "client_credentials grant."
            )
        return values

    def is_authenticated(self) -> bool:
        return True


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi


class IResponseHandler:
    __module__: str = 'cbra.types'

    async def on_response(
        self,
        request: fastapi.Request,
        response: fastapi.Response
    ) -> fastapi.Response | None:
        raise NotImplementedError


import typing

import pydantic

import cbra
import cbra.resource


class FooExampleResourceModel(pydantic.BaseModel):
    foo: int


class BarExampleResourceModel(pydantic.BaseModel):
    bar: int


class ExampleActionModel(pydantic.BaseModel):
    baz: int


class ExampleResource(cbra.resource.Resource):
    name: str = 'resource'
    name_article: str = 'an'
    pluralname: str = 'resources'
    path_parameter: str = 'id:int'
    model: typing.Any = BarExampleResourceModel | FooExampleResourceModel
    require_authentication: bool = False
    verbose_name: str = 'Example'

    filter_options = [
        cbra.resource.FilterOption(
            alias='search-param1',
            annotation=int,
            description="An integer search parameter"
        )
    ]


    query_parameters = [
        cbra.Query(
            alias="foo",
            default=None,
            annotation=int,
            description="The value of foo"
        ),
        cbra.Query(
            alias="bar",
            default=None,
            annotation=str
        ),
        cbra.Query(
            alias="baz",
            default=[],
            annotation=typing.Optional[typing.List[str]]
        )
    ]

    async def create(self, dto: model) -> FooExampleResourceModel:
        """Create a new resource."""
        return {
            'action': self.action,
            'path_params': self.get_path_params(),
            'dto': dto.dict()
        }

    async def list(self):
        return {
            'action': self.action,
            'path_params': self.get_path_params(),
            'query': self.get_query_parameters()
        }

    async def purge(self):
        return {'action': self.action, 'path_params': self.get_path_params()}

    async def retrieve(self, id: int):
        return {'action': self.action, 'id': id, 'path_params': self.get_path_params()}

    async def update(self, id: int):
        return {'action': self.action, 'id': id, 'path_params': self.get_path_params()}

    async def replace(self, id: int, dto: model):
        return {
            'action': self.action,
            'path_params': self.get_path_params(),
            'dto': dto.dict()
        }

    async def delete(self, id: int):
        return {'action': self.action, 'id': id, 'path_params': self.get_path_params()}


def get_asgi_application() -> cbra.Application:
    app = cbra.Application()
    app.add(ExampleResource, base_path='/')
    return app


app = get_asgi_application()
if __name__ == '__main__':
    cbra.run('__main__:app', reload=True, port=5000)



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime

import pydantic


class AuthorizedScope(pydantic.BaseModel):
    """Represents a scope that a Resource Owner has granted to a specific
    Client.
    """
    name: str = pydantic.Field(
        default=...
    )

    authorized: datetime.datetime = pydantic.Field(
        default_factory=datetime.datetime.utcnow
    )

    expires: datetime.datetime | None = pydantic.Field(
        default=None
    )

    def is_expired(self, now: datetime.datetime | None = None) -> bool:
        """Return a boolean indicating if the scope is expired."""
        now = now or datetime.datetime.utcnow()
        return self.expires is not None and (self.expires <= now)


"""Declares :class:`AuthorizationEndpointTest`."""
import pytest
from httpx import AsyncClient

from cbra.ext.oauth2 import AccessTokenSigner


@pytest.mark.asyncio
async def test_authentication_is_required(
    client: AsyncClient,
    path: str,
):
    response = await client.get(
        url=path,
    )
    assert response.status_code == 401, response.content


@pytest.mark.asyncio
async def test_authenticate_subject(
    base_url: str,
    client: AsyncClient,
    path: str,
    signer: AccessTokenSigner,
    signing_key: str
):
    at = await signer.sign(
        client_id='test',
        audience=base_url,
        sub=1,
        signing_keys=[signing_key]
    )
    response = await client.get(
        url=path,
        headers={'Authorization': f'Bearer {at}'}
    )
    assert response.status_code == 200, response.content


@pytest.mark.asyncio
async def test_invalid_token_type_is_rejected(
    base_url: str,
    client: AsyncClient,
    path: str,
    signer: AccessTokenSigner,
    signing_key: str
):
    at = await signer.sign(
        client_id='test',
        audience=base_url,
        sub=1,
        content_type="jwt",
        signing_keys=[signing_key]
    )
    response = await client.get(
        url=path,
        headers={'Authorization': f'Bearer {at}'}
    )
    assert response.status_code == 400, response.status_code


@pytest.mark.asyncio
async def test_expired_token_is_rejected(
    base_url: str,
    client: AsyncClient,
    path: str,
    signer: AccessTokenSigner,
    signing_key: str
):
    at = await signer.sign(
        client_id='test',
        audience=base_url,
        sub=1,
        signing_keys=[signing_key],
        now=1
    )
    response = await client.get(
        url=path,
        headers={'Authorization': f'Bearer {at}'}
    )
    assert response.status_code == 403, response.status_code


@pytest.mark.asyncio
async def test_invalid_audience_is_rejected(
    client: AsyncClient,
    path: str,
    signer: AccessTokenSigner,
    signing_key: str
):
    at = await signer.sign(
        client_id='test',
        audience="https://www.example.com",
        sub=1,
        signing_keys=[signing_key],
        now=1
    )
    response = await client.get(
        url=path,
        headers={'Authorization': f'Bearer {at}'}
    )
    assert response.status_code == 403, response.status_code


@pytest.mark.asyncio
async def test_unkown_signer_is_rejected_for_local_audience(
    base_url: str,
    client: AsyncClient,
    path: str,
    unknown_signer: AccessTokenSigner,
    signing_key: str
):
    at = await unknown_signer.sign(
        client_id='test',
        audience=base_url,
        sub=1,
        signing_keys=[signing_key],
    )
    response = await client.get(
        url=path,
        headers={'Authorization': f'Bearer {at}'}
    )
    assert response.status_code == 403, response.status_code


# pylint: skip-file
import functools
from typing import Any
from typing import Callable

from cbra.exceptions import MissingScope
from cbra.resource import Resource
from . import exceptions
from . import params
from . import types
from .accesstokensigner import AccessTokenSigner
from .authorization import Authorization
from .authorizeendpoint import AuthorizationEndpoint
from .authorizationserver import AuthorizationServer
from .authorizationrequestclient import AuthorizationRequestClient
from .clientconfig import ClientConfig
from .configfileclientrepository import ConfigFileClientRepository
from .introspectionendpoint import IntrospectionEndpoint
from .memorystorage import MemoryStorage
from .memoryclientrepository import MemoryClientRepository
from .memorysubjectrepository import MemorySubjectRepository
from .nullsubjectrepository import NullSubjectRepository
from .oidcclaimhandler import OIDCClaimHandler
from .oidctokenbuilder import OIDCTokenBuilder
from .pushedauthorizationrequestendpoint import PushedAuthorizationRequestEndpoint
from .rfc9068principal import RFC9068Principal
from .settingsclientrepository import SettingsClientRepository
from .settingssubjectrepository import SettingsSubjectRepository
from .staticsubjectepository import StaticSubjectRepository
from .tokenissuer import TokenIssuer
from .tokenrequesthandler import TokenRequestHandler
from .upstreamreturnhandler import UpstreamReturnHandler
from .upstreamprovider import UpstreamProvider


__all__: list[str] = [
    'exceptions',
    'params',
    'scope',
    'types',
    'AccessTokenSigner',
    'Authorization',
    'AuthorizationEndpoint',
    'AuthorizationServer',
    'AuthorizationRequestClient',
    'ClientConfig',
    'ConfigFileClientRepository',
    'IntrospectionEndpoint',
    'MemoryClientRepository',
    'MemoryStorage',
    'MemorySubjectRepository',
    'NullSubjectRepository',
    'OIDCClaimHandler',
    'OIDCTokenBuilder',
    'PushedAuthorizationRequestEndpoint',
    'RFC9068Principal',
    'SettingsClientRepository',
    'SettingsSubjectRepository',
    'StaticSubjectRepository',
    'TokenIssuer',
    'TokenRequestHandler',
    'UpstreamReturnHandler',
    'UpstreamProvider',
]


def scope(required: str | set[str]) -> Callable[..., Any]:
    """Method decorator for :class:`~cbra.Resource` implementations
    that enforces a given scope.
    """
    if isinstance(required, str):
        required = {required}

    def decorator_factory(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        async def f(self: Resource, *args: Any, **kwargs: Any):
            assert isinstance(required, set) # nosec
            if not self.principal.has_scope(required):
                missing: set[str] = required - self.principal.get_current_scope()
                raise MissingScope(missing=missing)
            return await func(self, *args, **kwargs)
        return f
    return decorator_factory


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`CloudSchedulerEndpoint`."""
from typing import Any

import aorta

from cbra.utils import classproperty
from .cloudschedulerendpoint import CloudSchedulerEndpoint


class CommandEndpoint(CloudSchedulerEndpoint):
    """A :class:`~cbra.ext.google.CloudSchedulerEndpoint` implementation that handles
    a request that is received from Google Cloud Scheduler and issues a command.
    """
    __module__: str = 'cbra.ext.google'
    default_response_code: int = 204
    response_description: str = "The command was issued."
    tags: list[str] = ["Beat"]

    @classproperty
    def responses(cls) -> dict[str | int, Any]:
        return {
            204: {'description': f"The command is issued."},
            401: {'description': "The request requires authentication."},
            403: {'description': "The principal is not allowed to beat the application."},
            422: {'description': "Parameters are required to invoke the command."}
        }

    @classmethod
    def as_command_handler(
        cls,
        name: str,
        command: type[aorta.Command],
        summary: str | None = None
    ) -> Any:
        """Create a new :class:`CommandEndpoint` implementation that
        handles the given command.
        """
        async def f(self: CommandEndpoint, dto: command._model): # type: ignore
            await self.publisher.publish(command(**dto.dict()))
        return cls.new(
            command_class=command,
            handle=f,
            methods=['POST'],
            tags=['Beats'],
            mount_path=f'.well-known/beat/{name}',
            name=f'commands.{name}',
            summary=summary or f"Issue {command.__name__}",
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic


class IntrospectionResponse(pydantic.BaseModel):
    active: bool = pydantic.Field(
        default=...,
        title="Active",
        description="Token active status"
    )

    sub: Any = pydantic.Field(
        default=None,
        title="Subject",
        description="The public identifier of the Subject."
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi

from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import JSONWebKeySet

from .iresponsehandler import IResponseHandler


class Request(fastapi.Request):

    @property
    def codec(self) -> PayloadCodec:
        """The :class:`~ckms.jose.PayloadCodec` instance used to
        decode JOSE objects.
        """
        return self.app.codec

    @property
    def keychain(self) -> Keychain:
        """The :class:`~ckms.core.Keychain` instance holding the
        signing and decryption keys used by the application.
        """
        return self.app.keychain

    @property
    def jwks(self) -> JSONWebKeySet:
        """A :class:`~ckms.types.JSONWebKeySet` holding the public keys intended
        for external consumers.
        """
        return self.keychain.tagged('unimatrixone.io/public-keys').as_jwks()

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.response_handlers = []

    def add_response_handler(self, handler: IResponseHandler):
        if 'handlers' not in self.scope:
            self.scope['handlers'] = []
        self.scope['handlers'].append(handler)

    async def run_response_handlers(
        self,
        response: fastapi.Response
    ) -> fastapi.Response | None:
        """Run all response handlers that are registered for this
        request.
        """
        for handler in self.response_handlers:
            result = await handler.on_response(self, response)
            if result is not None:
                response = result
                break
        return response




# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest
from ckms.core import Keychain

from cbra import Application
from cbra.ext.service import ServiceResource
from cbra.ext.oauth2 import AccessTokenSigner
from cbra.ext.oauth2.tests.authorization import *


@pytest.fixture
def app(server_keychain: Keychain) -> Application:
    app = Application(
        keychain=server_keychain
    )
    app.add(ObjectResource)
    return app


@pytest.fixture
def path() -> str:
    return '/objects'


@pytest.fixture
def signer(token_signer: AccessTokenSigner) -> AccessTokenSigner:
    return token_signer


@pytest.fixture
def unknown_signer(unknown_token_signer: AccessTokenSigner) -> AccessTokenSigner:
    return unknown_token_signer


class ObjectResource(ServiceResource):
    path_parameter: str = 'object_id'

    async def list(self):
        pass


# pylint: skip-file
from .accesstype import AccessType
from .authorizationcode import AuthorizationCode
from .authorizationcodegrant import AuthorizationCodeGrant
from .authorizationexception import AuthorizationException
from .authorizationidentifier import AuthorizationIdentifier
from .authorizationrequest import AuthorizationRequest
from .authorizationrequestclaims import AuthorizationRequestClaims
from .authorizationrequestparameters import AuthorizationRequestParameters
from .baseauthorizationrequest import BaseAuthorizationRequest
from .basegrant import BaseGrant
from .basesubject import BaseSubject
from .clientassertion import ClientAssertion
from .clientassertiontype import ClientAssertionType
from .clientcredentialsgrant import ClientCredentialsGrant
from .configurable import Configurable
from .dependantmodel import DependantModel
from .fields import GrantTypeField
from .fields import ResourceField
from .granttype import GrantType
from .iauthorization import IAuthorization
from .iauthorizeendpoint import IAuthorizeEndpoint
from .iauthorizationserver import IAuthorizationServer
from .iclient import IClient
from .iclientrepository import IClientRepository
from .introspectionrequest import IntrospectionRequest
from .introspectionresponse import IntrospectionResponse
from .iopenauthorizationserver import IOpenAuthorizationServer
from .iopenidtokenbuilder import IOpenIdTokenBuilder as IOIDCTokenBuilder
from .iprincipal import IPrincipal
from .irefreshtoken import IRefreshToken
from .istorage import IStorage
from .isubject import ISubject
from .isubjectrepository import ISubjectRepository
from .itokenissuer import ITokenIssuer
from .iupstreamprovider import IUpstreamProvider
from .iupstreamreturnhandler import IUpstreamReturnHandler
from .jar import JAR
from .jwtbearerassertiongrant import JWTBearerAssertionGrant
from .nullrefreshtoken import NullRefreshToken
from .oidcclaimrequest import OIDCClaimRequest
from .oidcrequestedclaims import OIDCRequestedClaims
from .policyfailure import PolicyFailure
from .promptexception import PromptException
from .redirecturl import RedirectURL
from .refreshtoken import RefreshToken
from .refreshtokengrant import RefreshTokenGrant
from .refreshtokenidentifier import RefreshTokenIdentifier
from .refreshtokenpolicy import RefreshTokenPolicy
from .responsemode import ResponseMode
from .responsetype import ResponseType
from .rfc9068token import RFC9068Token
from .scopedgrant import ScopedGrant
from .servermetadata import ServerMetadata
from .sessiongrant import SessionGrant
from .spaceseparatedlist import SpaceSeparatedList
from .stringorlist import StringOrList
from .tokenendpointresponse import TokenEndpointResponse
from .tokenexception import TokenException
from .tokenrequestparameters import TokenRequestParameters
from .tokenresponse import TokenResponse
from .tokentype import TokenType


__all__: list[str] = [
    'AccessType',
    'AuthorizationCode',
    'AuthorizationCodeGrant',
    'AuthorizationIdentifier',
    'AuthorizationException',
    'AuthorizationRequest',
    'AuthorizationRequestClaims',
    'AuthorizationRequestParameters',
    'BaseAuthorizationRequest',
    'BaseGrant',
    'BaseSubject',
    'ClientAssertion',
    'ClientAssertionType',
    'ClientCredentialsGrant',
    'Configurable',
    'DependantModel',
    'GrantType',
    'GrantTypeField',
    'IAuthorization',
    'IAuthorizeEndpoint',
    'IAuthorizationServer',
    'IClient',
    'IClientRepository',
    'IntrospectionRequest',
    'IntrospectionResponse',
    'IOpenAuthorizationServer',
    'IOIDCTokenBuilder',
    'IPrincipal',
    'IRefreshToken',
    'IStorage',
    'ISubject',
    'ISubjectRepository',
    'ITokenIssuer',
    'IUpstreamProvider',
    'IUpstreamReturnHandler',
    'JAR',
    'JWTBearerAssertionGrant',
    'NullRefreshToken',
    'OIDCClaimRequest',
    'OIDCRequestedClaims',
    'PromptException',
    'PolicyFailure',
    'RedirectURL',
    'RefreshToken',
    'RefreshTokenGrant',
    'RefreshTokenIdentifier',
    'RefreshTokenPolicy',
    'ResourceField',
    'ResponseType',
    'ResponseMode',
    'RFC9068Token',
    'ScopedGrant',
    'ServerMetadata',
    'SessionGrant',
    'SpaceSeparatedList',
    'StringOrList',
    'TokenEndpointResponse',
    'TokenException',
    'TokenRequestParameters',
    'TokenResponse',
    'TokenType',
]


"""Declares :class:`GunicornServer`."""
import multiprocessing
import typing

import fastapi
import gunicorn.app.base
from unimatrix.conf import settings


class GunicornServer(gunicorn.app.base.BaseApplication):

    @staticmethod
    def number_of_workers():
        return (multiprocessing.cpu_count() * 2) + 1

    def __init__(self,
        app: fastapi.FastAPI,
        options: typing.Optional[dict] = None
    ):
        self.application = app
        self.options = options or {}
        self.options.update({
            'workers': settings.WEB_CONCURRENCY\
                or (self.number_of_workers() if not settings.DEBUG else 1)
        })
        self.options.setdefault('timeout', settings.HTTP_WORKER_TIMEOUT)
        self.options.setdefault('worker_class', 'uvicorn.workers.UvicornWorker')
        if settings.DEPLOYMENT_ENV == 'local':
            self.configure_local()

        super().__init__()

    def configure_local(self) -> None:
        """Configures the server to run in a local development environment."""
        if settings.LOCALHOST_SSL_CRT and settings.LOCALHOST_SSL_KEY:
            self.options.update({
                'certfile': settings.LOCALHOST_SSL_CRT,
                'keyfile': settings.LOCALHOST_SSL_KEY
            })

    def load_config(self) -> None:
        """Constructs the configuration from the parameters provided to the
        constructor.
        """
        config = {
            key: value for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self) -> fastapi.FastAPI:
        """Return the ASGI application handling incoming requests."""
        return self.application




# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cbra.types import Request
from cbra.types import IParser
from ..exceptions import ParseError


class JWTParser(IParser):
    """Parses the JSON Web Encryption (JWE) or JSON Web Signature (JWS)
    from the request body and decodes the payload as a JSON Web
    Token (JWT).
    """
    media_type: str = "application/jwt"
    token_types: set[str] = {"jwt"}

    async def parse(
        self,
        request: Request,
        media_type: str | None = None,
        parser_context: dict[str, Any] | None = None
    ) -> Any:
        """Parse the request body to a JSON Web Signature (JWS) and
        return the payload as a :class:`~ckms.types.JSONWebToken`
        instance. The signature of the JWS and the claims of the JWT
        are not verified and validated.
        
        If the request body is JSON Web Encryption (JWE), the :meth:`parse()`
        method uses the :attr:`~cbra.Application.keychain` to decrypt the
        payload.

        A :exc:`~cbra.exceptions.ParseError` is raised if the JOSE object in
        the request body is malformed, is not a JWS or if the payload of the
        JWS is not a JWT.

        The following failure status may be produced by :meth:`parse()`:

        - `MALFORMED_BODY` - The request body could not be decrypted,
          or the decrypted object could not be parsed as a JWS/JWT.
        - `INVALID_TOKEN_TYPE` - The request body was succesfully decrypted
          and the JOSE header could be interpreted, but its ``typ`` claim
          specified a value that is not accepted by the parser.
        """
        try:
            jws = await request.codec.jws(await request.body())
        except Exception:
            raise ParseError(
                code="MALFORMED_BODY",
                message=(
                    "The request body could not be parsed as a JSON "
                    "Web Signature (JWS)."
                )
            )
        if not jws.is_jwt(self.token_types):
            raise ParseError(
                code="INVALID_TOKEN_TYPE",
                message=(
                    "The payload of the JSON Web Signature (JWS) is not a JSON "
                    "Web Token (JWT) or the JOSE header declared a 'typ' claim "
                    "that is not accepted by the server."
                )
            )
        return jws


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
from typing import Any

from cbra.ext.api.credential import ApplicationClientCredential
from cbra.ext.api import Consumer


class TestConsumer(Consumer['TestConsumer']):

    def get_http_client_kwargs(self) -> dict[str, Any]:
        return {}


async def main():
    credential = ApplicationClientCredential()
    client = TestConsumer()
    await client.request(
        method='GET',
        response_model=lambda x: x,
        url='https://www.example.com',
        credential=credential
    )


if __name__ == '__main__':
    asyncio.run(main())

